@@ -1,9 +1,10 @@
 import binascii
 import enum
 import logging
+import warnings
 from dataclasses import dataclass
 from datetime import timedelta
-from typing import Dict, Iterable, List, Optional, Union
+from typing import Any, Dict, Iterable, List, Optional, Set, Union
 
 import yaml
 from asn1crypto import x509
@@ -370,6 +371,84 @@ class PemDerSignatureConfig(config_utils.ConfigurableMixin):
         return result
 
 
+@dataclass(frozen=True)
+class TokenCriteria(config_utils.ConfigurableMixin):
+    """
+    .. versionadded:: 0.14.0
+
+    Search criteria for a PKCS#11 token.
+    """
+
+    label: Optional[str] = None
+    """
+    Label of the token to use. If ``None``, there is no constraint.
+    """
+
+    serial: Optional[bytes] = None
+    """
+    Serial number of the token to use. If ``None``, there is no constraint.
+    """
+
+    @classmethod
+    def process_entries(cls, config_dict):
+        try:
+            config_dict['serial'] = binascii.unhexlify(config_dict['serial'])
+        except KeyError:
+            pass
+        except ValueError as e:
+            raise ConfigurationError(
+                "Failed to parse PKCS #11 token serial number as a hex string"
+            ) from e
+
+
+class PKCS11PinEntryMode(enum.Enum):
+    """
+    Pin entry behaviour if the user PIN is not supplied as part of the config.
+    """
+
+    PROMPT = enum.auto()
+    """
+    Prompt for a PIN (the default).
+
+    .. note::
+        This value is only processed by the CLI, and ignored when the PKCS#11
+        signer is called from library code. In those cases, the default
+        is effectively :attr:`SKIP`.
+    """
+
+    DEFER = enum.auto()
+    """
+    Let the PKCS #11 module handle its own authentication during login.
+
+    .. note::
+        This applies to some devices that have physical PIN pads, for example.
+    """
+
+    SKIP = enum.auto()
+    """
+    Skip the login process altogether.
+
+    .. note::
+        This applies to some devices that manage user authentication outside
+        the scope of PKCS #11 entirely.
+    """
+
+    @staticmethod
+    def parse_mode_setting(value: Any) -> 'PKCS11PinEntryMode':
+        if isinstance(value, str):
+            try:
+                return PKCS11PinEntryMode.__members__[value.upper()]
+            except KeyError:
+                raise ConfigurationError(
+                    f"Invalid PIN entry mode {value!r}; must be one of "
+                    f"{', '.join(repr(x.name) for x in PKCS11PinEntryMode)}."
+                )
+        else:
+            # fallback for backwards compatibility
+            return PKCS11PinEntryMode.PROMPT if value \
+                else PKCS11PinEntryMode.SKIP
+
+
 @dataclass(frozen=True)
 class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
     """
@@ -400,10 +479,10 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
         the one provided on the token.
     """
 
-    token_label: Optional[str] = None
+    token_criteria: Optional[TokenCriteria] = None
     """PKCS#11 token name"""
 
-    other_certs: List[x509.Certificate] = None
+    other_certs: Optional[List[x509.Certificate]] = None
     """Other relevant certificates."""
 
     key_label: Optional[str] = None
@@ -435,9 +514,9 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
         set :attr:`prompt_pin` to ``False``.
     """
 
-    prompt_pin: bool = True
+    prompt_pin: PKCS11PinEntryMode = PKCS11PinEntryMode.PROMPT
     """
-    Prompt for the user's PIN. Default is ``True``.
+    Set PIN entry and PKCS #11 login behaviour.
 
     .. note::
         If :attr:`user_pin` is not ``None``, this setting has no effect.
@@ -471,6 +550,15 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
         This is currently only supported for ECDSA signatures.
     """
 
+    @classmethod
+    def check_config_keys(cls, keys_supplied: Set[str]):
+        # make sure we don't ding token_label since we actually still
+        # process it for compatibility reasons
+        super().check_config_keys({
+            k for k in keys_supplied
+            if k not in ('token_label', 'token-label')
+        })
+
     @classmethod
     def process_entries(cls, config_dict):
         super().process_entries(config_dict)
@@ -479,12 +567,6 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
             other_certs = (other_certs,)
         config_dict['other_certs'] = list(load_certs_from_pemder(other_certs))
 
-        if 'token_label' not in config_dict and 'slot_no' not in config_dict:
-            raise ConfigurationError(
-                "Either 'slot_no' or 'token_label' must be provided in "
-                "PKCS#11 setup"
-            )
-
         cert_file = config_dict.get('signing_certificate', None)
         if cert_file is not None:
             config_dict['signing_certificate'] \
@@ -509,6 +591,23 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
                 "must be provided in PKCS#11 setup"
             )
 
+        config_dict['prompt_pin'] = get_and_apply(
+            config_dict, 'prompt_pin', PKCS11PinEntryMode.parse_mode_setting,
+            default=PKCS11PinEntryMode.PROMPT
+        )
+
+        if 'token_label' in config_dict:
+            warnings.warn(
+                "'token_label' is deprecated, use 'token_criteria.label' "
+                "instead",
+                DeprecationWarning
+            )
+            lbl = config_dict.pop('token_label')
+            if 'token_criteria' not in config_dict:
+                config_dict['token_criteria'] = {'label': lbl}
+            else:
+                config_dict['token_criteria'].setdefault('label', lbl)
+
 
 def _process_pkcs11_id_value(x: Union[str, int]):
     if isinstance(x, int):
