@@ -12,6 +12,7 @@ from typing import List, Optional
 from asn1crypto import x509
 
 from . import crypt, generic, misc, writer
+from .font.basic import get_courier
 from .generic import pdf_name, pdf_string
 
 __all__ = [
@@ -19,8 +20,6 @@ __all__ = [
     'FileSpec', 'RelatedFileSpec', 'wrap_encrypted_payload'
 ]
 
-from .misc import get_courier
-
 
 @dataclass(frozen=True)
 class EmbeddedFileParams:
@@ -460,7 +459,7 @@ def wrap_encrypted_payload(plaintext_payload: bytes, *,
     if include_explanation_page:
         resources = generic.DictionaryObject({
             pdf_name('/Font'): generic.DictionaryObject({
-                pdf_name('/F1'): get_courier()
+                pdf_name('/F1'): get_courier(w)
             })
         })
 
