@@ -8,6 +8,7 @@ of the PyPDF2 project.
 import binascii
 import codecs
 import decimal
+import enum
 import logging
 import os
 import re
@@ -19,13 +20,15 @@ from typing import Any, Callable, Iterator, Optional, Tuple, Union
 from . import filters
 from .misc import (
     IndirectObjectExpected,
+    PdfError,
     PdfReadError,
     PdfStreamError,
+    PdfStrictReadError,
     PdfWriteError,
     is_regular_character,
     read_non_whitespace,
     read_until_regex,
-    skip_over_comment,
+    skip_over_whitespace,
 )
 
 __all__ = [
@@ -33,7 +36,8 @@ __all__ = [
     'PdfObject', 'IndirectObject', 'NullObject', 'BooleanObject', 'FloatObject',
     'NumberObject', 'ByteStringObject', 'TextStringObject', 'NameObject',
     'ArrayObject', 'DictionaryObject', 'StreamObject',
-    'read_object', 'pdf_name', 'pdf_string', 'pdf_date'
+    'read_object', 'pdf_name', 'pdf_string', 'pdf_date',
+    'TextStringEncoding', 'EncryptedObjAccess', 'DecryptedObjectProxy'
 ]
 
 OBJECT_PREFIXES = b'/<[tf(n%'
@@ -43,15 +47,49 @@ INDIRECT_PATTERN = re.compile(r"(\d+)\s+(\d+)\s+R[^a-zA-Z]".encode('ascii'))
 logger = logging.getLogger(__name__)
 
 
-def _deproxy_decrypt(obj):
+class EncryptedObjAccess(enum.Enum):
+    """
+    Defines what to do when an encrypted object is encountered when retrieving
+    an object from a container.
+    """
+
+    PROXY = 0
+    """
+    Return the proxy object as-is, and leave further encryption/decryption
+    handling to the caller.
+    """
+
+    TRANSPARENT = 1
+    """
+    Transparently decrypt the proxy's content (similarly wrapping any
+    sub-containers in :class:`.DecryptedObjectProxy`, so this applies
+    recursively).
+
+    .. note::
+        This is the default in most situations, since it's the least likely
+        to get in the way of any APIs that are not explicitly aware of
+        content encryption concerns.
+    """
+
+    RAW = 2
+    """
+    Return the underlying raw object as written, without attempting or deferring
+    decryption.
+    """
+
+
+def _deproxy_decrypt(obj, eoa: EncryptedObjAccess):
     if isinstance(obj, DecryptedObjectProxy):
-        return obj.decrypted
-    else:
-        return obj
+        if eoa == EncryptedObjAccess.TRANSPARENT:
+            return obj.decrypted
+        elif eoa == EncryptedObjAccess.RAW:
+            return obj.raw_object
+    return obj
 
 
 class Dereferenceable:
-    """Represents an opaque reference to a PDF object associated with
+    """
+    Represents an opaque reference to a PDF object associated with
     a PDF Handler (see :class:`PdfHandler <.rw_common.PdfHandler>`).
 
     This can either be a reference to an object with an object ID
@@ -75,20 +113,19 @@ class Dereferenceable:
 
 
 class TrailerReference(Dereferenceable):
-    """A reference to the trailer of a PDF document.
+    """
+    A reference to the trailer of a PDF document.
 
     .. warning::
        Since the trailer does not have a well-defined object ID in files with
        "classical" cross-reference tables (as opposed to cross-reference
        streams), this is not a subclass of :class:`.Reference`.
+
+    :param reader:
+        a :class:`~pyhanko.pdf_utils.reader.PdfFileReader`
     """
 
     def __init__(self, reader):
-        """Create a reference to the trailer of a :class:`.reader.PdfFileReader`
-        instance.
-
-        :param reader: A PDF reader
-        """
         self.reader = reader
 
     def get_object(self) -> 'PdfObject':
@@ -100,7 +137,8 @@ class TrailerReference(Dereferenceable):
 
 @dataclass(frozen=True)
 class Reference(Dereferenceable):
-    """A reference to an object with a certain ID and generation number, with
+    """
+    A reference to an object with a certain ID and generation number, with
     a PDF handler attached to it.
 
     .. warning::
@@ -147,8 +185,10 @@ class Reference(Dereferenceable):
         return self.pdf
 
 
-def read_object(stream, container_ref: 'Dereferenceable') -> 'PdfObject':
-    """Read a PDF object from an input stream.
+def read_object(stream, container_ref: 'Dereferenceable',
+                as_metadata_stream: bool = False) -> 'PdfObject':
+    """
+    Read a PDF object from an input stream.
 
     .. note::
        The `container_ref` parameter tells the API which reference to register
@@ -163,6 +203,8 @@ def read_object(stream, container_ref: 'Dereferenceable') -> 'PdfObject':
 
         *Note:* It is perfectly possible (and common) for `container_ref` to
         resolve to the return value of this function.
+    :param as_metadata_stream:
+        Whether to dereference the object as an XMP metadata stream.
     :return:
         A :class:`.PdfObject`.
     """
@@ -179,7 +221,7 @@ def read_object(stream, container_ref: 'Dereferenceable') -> 'PdfObject':
         stream.seek(-2, os.SEEK_CUR)  # reset to start
         if peek == b'<<':
             result = DictionaryObject.read_from_stream(
-                stream, container_ref
+                stream, container_ref, as_metadata_stream=as_metadata_stream
             )
         else:
             result = read_hex_string_from_stream(stream)
@@ -258,7 +300,7 @@ class PdfObject:
         """
         ref = self.container_ref
         if ref is None:  # pragma: nocover
-            raise ValueError(
+            raise PdfReadError(
                 'No container reference available. This object probably '
                 'wasn\'t read from a file.'
             )
@@ -287,7 +329,8 @@ class PdfObject:
 
 
 class NullObject(PdfObject):
-    """PDF `null` object.
+    """
+    PDF `null` object.
 
     All instances are treated as equal and falsy.
     """
@@ -350,7 +393,8 @@ class BooleanObject(PdfObject):
 
 
 class ArrayObject(list, PdfObject):
-    """PDF array object. This class extends from Python's list class,
+    """
+    PDF array object. This class extends from Python's list class,
     and supports its interface.
 
     .. warning::
@@ -369,8 +413,13 @@ class ArrayObject(list, PdfObject):
     def __getitem__(self, index):
         return self.raw_get(index).get_object()
 
-    def raw_get(self, index, decrypt=True):
+    def raw_get(self, index,
+                decrypt: EncryptedObjAccess = EncryptedObjAccess.TRANSPARENT):
         """
+        .. versionchanged:: 0.14.0
+
+            ``decrypt`` parameter is no longer boolean
+
         Get a value from an array without dereferencing.
         In other words, if the value corresponding to the given key is of type
         :class:`.IndirectObject`, the indirect reference will not be resolved.
@@ -378,15 +427,15 @@ class ArrayObject(list, PdfObject):
         :param index:
             Key to look up in the dictionary.
         :param decrypt:
-            If ``False``, instances of :class:`.DecryptedObjectProxy` will
-            be returned as-is. If ``True``, they will be decrypted.
-            Default ``True``.
+            What to do when retrieving encrypted objects; see
+            :class:`.EncryptedObjAccess`. The default is
+            :attr:`.EncryptedObjAccess.TRANSPARENT`.
         :return:
             A :class:`.PdfObject`.
         """
 
         val = list.__getitem__(self, index)
-        return _deproxy_decrypt(val) if decrypt else val
+        return _deproxy_decrypt(val, decrypt)
 
     def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(b"[")
@@ -514,7 +563,8 @@ class IndirectObject(PdfObject, Dereferenceable):
 
 
 class FloatObject(decimal.Decimal, PdfObject):
-    """PDF Float object.
+    """
+    PDF Float object.
 
     Internally, these are treated as decimals (and therefore actually
     fixed-point objects, to be precise).
@@ -569,7 +619,11 @@ class NumberObject(int, PdfObject):
 
     @staticmethod
     def read_from_stream(stream):
-        num = read_until_regex(stream, NumberObject.NumberPattern)
+        num = read_until_regex(
+            stream, regex=NumberObject.NumberPattern,
+            # for consistency with other read_object() output
+            ignore_eof=True
+        )
         if num.find(NumberObject.ByteDot) != -1:
             return FloatObject(num.decode('ascii'))
         else:
@@ -578,7 +632,8 @@ class NumberObject(int, PdfObject):
 
 # TODO: not sure I like this behaviour of PyPDF2. Review.
 
-def pdf_string(string) -> Union['ByteStringObject', 'TextStringObject']:
+def pdf_string(string: Union[str, bytes, bytearray]) \
+        -> Union['ByteStringObject', 'TextStringObject']:
     """
     Encode a string as a :class:`.TextStringObject` if possible,
     or a :class:`.ByteStringObject` otherwise.
@@ -589,19 +644,11 @@ def pdf_string(string) -> Union['ByteStringObject', 'TextStringObject']:
     if isinstance(string, str):
         return TextStringObject(string)
     elif isinstance(string, (bytes, bytearray)):
+        guessed = _guess_enc_by_bom(string)
         try:
-            if string.startswith(codecs.BOM_UTF16_BE):
-                retval = TextStringObject(string.decode("utf-16"))
-                retval.autodetect_utf16 = True
-                return retval
-            else:
-                # This is probably a big performance hit here, but we need to
-                # convert string objects into the text/unicode-aware version if
-                # possible... and the only way to check if that's possible is
-                # to try.  Some strings are strings, some are just byte arrays.
-                retval = TextStringObject(decode_pdfdocencoding(string))
-                retval.autodetect_pdfdocencoding = True
-                return retval
+            retval = TextStringObject(guessed.decode(string))
+            retval.autodetected_encoding = guessed
+            return retval
         except UnicodeDecodeError:
             return ByteStringObject(string)
     else:
@@ -627,10 +674,7 @@ def read_hex_string_from_stream(stream) \
         nonlocal odd
         while True:
             tok = read_non_whitespace(stream)
-            if not tok:
-                # stream has truncated prematurely
-                raise PdfStreamError("Stream has ended unexpectedly")
-            elif tok == b">":
+            if tok == b">":
                 return
             elif tok not in HEX_DIGITS:
                 raise PdfStreamError(
@@ -644,21 +688,10 @@ def read_hex_string_from_stream(stream) \
     return pdf_string(result)
 
 
-def read_string_from_stream(stream) -> Union['ByteStringObject',
-                                             'TextStringObject']:
-    """
-    Read a PDF string literal from a stream.
-
-    This method should in principle always return a :class:`TextStringObject`
-    instance if the underlying file is a valid PDF file.
-
-    :param stream:
-        An input stream.
-    """
-
+def _read_string_literal_bytes(stream) -> bytes:
     stream.read(1)
     parens = 1
-    txt = b""
+    txt = BytesIO()
     while True:
         tok = stream.read(1)
         if not tok:
@@ -695,6 +728,8 @@ def read_string_from_stream(stream) -> Union['ByteStringObject',
                     if ntok.isdigit():
                         tok += ntok
                     else:
+                        # premature end, seek back
+                        stream.seek(-1, os.SEEK_CUR)
                         break
                 octal = int(tok, base=8)
                 # interpret as byte
@@ -710,14 +745,27 @@ def read_string_from_stream(stream) -> Union['ByteStringObject',
                 # line break was escaped:
                 tok = b''
             else:
-                raise PdfReadError(r"Unexpected escaped string: %s" % tok)
-        txt += tok
-    return pdf_string(txt)
+                raise PdfReadError("Unexpected escaped string: " + repr(tok))
+        txt.write(tok)
+    return txt.getvalue()
+
+
+def read_string_from_stream(stream) -> Union['ByteStringObject',
+                                             'TextStringObject']:
+    """
+    Read a PDF string literal from a stream. Attempt to decode it into a text
+    string by autodetecting the encoding, or failing that, return it as a byte
+    string instead.
+
+    :param stream:
+        An input stream.
+    """
+
+    return pdf_string(_read_string_literal_bytes(stream))
 
 
 class ByteStringObject(bytes, PdfObject):
-    """ PDF bytestring class.
-    """
+    """PDF bytestring class."""
 
     original_bytes = property(lambda self: self)
     """
@@ -737,19 +785,100 @@ class ByteStringObject(bytes, PdfObject):
         stream.write(b">")
 
 
+class TextStringEncoding(enum.Enum):
+    """
+    Encodings for PDF text strings.
+    """
+
+    PDF_DOC = None
+    """
+    PDFDocEncoding (one-byte character codes; PDF-specific).
+    """
+
+    UTF16BE = (codecs.BOM_UTF16_BE, 'utf-16be')
+    """
+    UTF-16BE encoding.
+    """
+
+    UTF8 = (codecs.BOM_UTF8, 'utf-8')
+    """
+    UTF-8 encoding (PDF 2.0)
+    """
+
+    UTF16LE = (codecs.BOM_UTF16_LE, 'utf-16le')
+    """
+    UTF-16LE encoding.
+
+    .. note::
+        This is strictly speaking invalid in PDF 2.0, but some authoring tools
+        output such strings anyway (presumably due to the fact that it's the
+        default wide character encoding on Windows).
+    """
+
+    def encode(self, string: str) -> bytes:
+        """
+        Encode a string with BOM.
+
+        :param string:
+            The string to encode.
+        :return:
+            The encoded string.
+        """
+        if self == TextStringEncoding.PDF_DOC:
+            return encode_pdfdocencoding(string)
+        else:
+            bom, enc = self.value
+            return bom + string.encode(enc)
+
+    def decode(self, string: Union[bytes, bytearray]) -> str:
+        """
+        Decode a string with BOM.
+
+        :param string:
+            The string to encode.
+        :return:
+            The encoded string.
+        :raise UnicodeDecodeError:
+            Raised if decoding fails.
+        """
+        if self == TextStringEncoding.PDF_DOC:
+            return decode_pdfdocencoding(string)
+        elif self == TextStringEncoding.UTF8:
+            return string.decode('utf-8-sig')
+        else:
+            return string.decode('utf-16')
+
+
+def _guess_enc_by_bom(encoded: Union[bytes, bytearray]) \
+        -> TextStringEncoding:
+    if encoded.startswith(codecs.BOM_UTF16_BE):
+        return TextStringEncoding.UTF16BE
+    elif encoded.startswith(codecs.BOM_UTF16_LE):
+        return TextStringEncoding.UTF16LE
+    elif encoded.startswith(codecs.BOM_UTF8):
+        return TextStringEncoding.UTF8
+    else:
+        # This is probably a big performance hit here, but we need to
+        # convert string objects into the text/unicode-aware version if
+        # possible... and the only way to check if that's possible is
+        # to try.  Some strings are strings, some are just byte arrays.
+        return TextStringEncoding.PDF_DOC
+
+
 class TextStringObject(str, PdfObject):
     """
     PDF text string object.
     """
 
-    autodetect_pdfdocencoding = False
+    autodetected_encoding: Optional[TextStringEncoding] = None
     """
-    If ``True``, this string was determined to be encoded in PDFDoc encoding.
+    Autodetected encoding when parsing the file.
     """
 
-    autodetect_utf16 = False
+    force_output_encoding: Optional[TextStringEncoding] = None
     """
-    If ``True``, this string was determined to be encoded in UTF16-BE encoding.
+    Output encoding to use when serialising the string.
+    The default is to try PDFDocEncoding first, and fall back to UTF-16BE.
     """
 
     @property
@@ -766,22 +895,25 @@ class TextStringObject(str, PdfObject):
         # we were wrong.  It's pretty common.  Return the original bytes that
         # would have been used to create this object, based upon the autodetect
         # method.
-        if self.autodetect_utf16:
-            return codecs.BOM_UTF16_BE + self.encode("utf-16be")
-        elif self.autodetect_pdfdocencoding:
-            return encode_pdfdocencoding(self)
+        if self.autodetected_encoding:
+            return self.autodetected_encoding.encode(self)
         else:
-            raise Exception("no information about original bytes")
+            raise PdfError("No information about original bytes")
 
     def write_to_stream(self, stream, handler=None, container_ref=None):
-        # Try to write the string out as a PDFDocEncoding encoded string.  It's
-        # nicer to look at in the PDF file.  Sadly, we take a performance hit
-        # here for trying...
-        bytearr: bytes
-        try:
-            bytearr = encode_pdfdocencoding(self)
-        except UnicodeEncodeError:
-            bytearr = codecs.BOM_UTF16_BE + self.encode("utf-16be")
+        if self.force_output_encoding is not None:
+            encoded = self.force_output_encoding.encode(self)
+        else:
+            # Try to write the string out as a PDFDocEncoding encoded string.
+            # It's nicer to look at in the PDF file.  Sadly, we take a
+            # performance hit here for trying...
+            encoded: bytes
+            try:
+                encoded = encode_pdfdocencoding(self)
+            except UnicodeEncodeError:
+                # fall back to UTF-16BE by default, since it's the only
+                # valid pre-2.0 Unicode encoding.
+                encoded = codecs.BOM_UTF16_BE + self.encode("utf-16be")
 
         cf = None
         if handler is not None and container_ref is not None:
@@ -793,12 +925,12 @@ class TextStringObject(str, PdfObject):
             local_key = cf.derive_object_key(
                 container_ref.idnum, container_ref.generation
             )
-            bytearr = cf.encrypt(local_key, bytearr)
-            obj = ByteStringObject(bytearr)
+            encoded = cf.encrypt(local_key, encoded)
+            obj = ByteStringObject(encoded)
             obj.write_to_stream(stream)
         else:
             stream.write(b"(")
-            for c in bytearr:
+            for c in encoded:
                 c_ = bytes([c])
                 if not c_.isalnum() and c != 0x20:
                     stream.write(b"\\%03o" % c)
@@ -845,7 +977,7 @@ def _decode_name(name_bytes: bytes) -> 'NameObject':
             elif not (0x21 <= cur_byte <= 0x7e) or \
                     not is_regular_character(cur_byte):
                 raise PdfReadError(
-                    f"Byte (0x{hex(cur_byte)}) must be escaped in a PDF name"
+                    f"Byte (0x{cur_byte:02x}) must be escaped in a PDF name"
                 )
             result.write(
                 bytes((cur_byte,))
@@ -882,7 +1014,9 @@ class NameObject(str, PdfObject):
     required.
     """
 
-    DELIMITER_PATTERN = re.compile(r"\s+|[\(\)<>\[\]{}/%]".encode('ascii'))
+    DELIMITER_PATTERN = re.compile(
+        r"\s|[\(\)<>\[\]{}/%]".encode('ascii') + b'|\x00'
+    )
 
     def write_to_stream(self, stream, handler=None, container_ref=None):
         byte_iter = iter(self.encode('utf8'))
@@ -939,8 +1073,13 @@ class DictionaryObject(dict, PdfObject):
         else:
             super().__init__()
 
-    def raw_get(self, key, decrypt=True):
+    def raw_get(self, key,
+                decrypt: EncryptedObjAccess = EncryptedObjAccess.TRANSPARENT):
         """
+        .. versionchanged:: 0.14.0
+
+            ``decrypt`` parameter is no longer boolean
+
         Get a value from a dictionary without dereferencing.
         In other words, if the value corresponding to the given key is of type
         :class:`.IndirectObject`, the indirect reference will not be resolved.
@@ -948,14 +1087,14 @@ class DictionaryObject(dict, PdfObject):
         :param key:
             Key to look up in the dictionary.
         :param decrypt:
-            If ``False``, instances of :class:`.DecryptedObjectProxy` will
-            be returned as-is. If ``True``, they will be decrypted.
-            Default ``True``.
+            What to do when retrieving encrypted objects; see
+            :class:`.EncryptedObjAccess`. The default is
+            :attr:`.EncryptedObjAccess.TRANSPARENT`.
         :return:
             A :class:`.PdfObject`.
         """
         val = dict.__getitem__(self, key)
-        return _deproxy_decrypt(val) if decrypt else val
+        return _deproxy_decrypt(val, decrypt)
 
     def __setitem__(self, key, value):
         key = _normalise_key(key)
@@ -974,7 +1113,16 @@ class DictionaryObject(dict, PdfObject):
         return dict.setdefault(self, key, value)
 
     def __getitem__(self, key):
-        return dict.__getitem__(self, key).get_object()
+        raw_obj = dict.__getitem__(self, key)
+        if key == '/Metadata' and isinstance(raw_obj, IndirectObject):
+            from pyhanko.pdf_utils.rw_common import PdfHandler
+            handler = raw_obj.get_pdf_handler()
+            assert isinstance(handler, PdfHandler)
+            return handler.get_object(
+                raw_obj.reference, as_metadata_stream=True
+            )
+        else:
+            return raw_obj.get_object()
 
     def get_and_apply(self, key, function: Callable[[PdfObject], Any], *,
                       raw=False, default=None):
@@ -1005,7 +1153,8 @@ class DictionaryObject(dict, PdfObject):
         stream.write(b">>")
 
     @staticmethod
-    def read_from_stream(stream, container_ref: 'Dereferenceable'):
+    def read_from_stream(stream, container_ref: 'Dereferenceable',
+                         as_metadata_stream: bool = False):
         tmp = stream.read(2)
         if tmp != b"<<":
             raise PdfReadError(
@@ -1016,16 +1165,6 @@ class DictionaryObject(dict, PdfObject):
         handler = container_ref.get_pdf_handler()
         while True:
             tok = read_non_whitespace(stream)
-            if tok == b'\x00':
-                continue
-            elif tok == b'%':
-                stream.seek(-1, os.SEEK_CUR)
-                skip_over_comment(stream)
-                continue
-            if not tok:
-                # stream has truncated prematurely
-                raise PdfStreamError("Stream has ended unexpectedly")
-
             if tok == b">":
                 stream.read(1)
                 break
@@ -1042,7 +1181,7 @@ class DictionaryObject(dict, PdfObject):
                     "%s for key %s" % (hex(stream.tell()), key)
                 )
                 if handler.strict:
-                    raise PdfReadError(err)
+                    raise PdfStrictReadError(err)
                 else:
                     logger.warning(err)
 
@@ -1050,16 +1189,10 @@ class DictionaryObject(dict, PdfObject):
         s = read_non_whitespace(stream, allow_eof=True)
         stream_data = None
         if s == b's' and stream.read(5) == b'tream':
-            eol = stream.read(1)
             # odd PDF file output has spaces after 'stream' keyword
-            # but before EOL. patch provided by Danial Sandler
-            while eol == b' ':
-                eol = stream.read(1)
-            assert eol in (b"\n", b"\r")
-            if eol == b"\r":
-                # read \n after
-                if stream.read(1) != b'\n':
-                    stream.seek(-1, os.SEEK_CUR)
+            # but before EOL. Original PyPDF2 patch provided by Danial Sandler,
+            # modified by Matthias Valvekens
+            skip_over_whitespace(stream, stop_after_eol=True)
             # this is a stream object, not a dictionary
             length = data[pdf_name("/Length")]
             if isinstance(length, IndirectObject):
@@ -1076,30 +1209,40 @@ class DictionaryObject(dict, PdfObject):
                 # and Python users into PDF files tend to be our audience.
                 # we need to do this to correct the streamdata and chop off
                 # an extra character.
-                pos = stream.tell()
+                orig_endstream_pos = stream.tell()
                 stream.seek(-10, os.SEEK_CUR)
                 end = stream.read(9)
                 if end == b"endstream":
                     # we found it by looking back one character further.
                     stream_data = stream_data[:-1]
                 else:
-                    stream.seek(pos)
                     raise PdfReadError(
                         "Unable to find 'endstream' marker after "
-                        "stream at byte %s." % hex(stream.tell())
+                        "stream at byte %s." % hex(orig_endstream_pos)
                     )
         else:
             stream.seek(pos)
         if stream_data is not None:
             # pass in everything as encoded data, the StreamObject class
             # will take care of decoding as necessary
-            return StreamObject(data, encoded_data=stream_data)
+            stm_cls = StreamObject
+            if as_metadata_stream:
+                try:
+                    # noinspection PyUnresolvedReferences
+                    from pyhanko.pdf_utils.metadata.xmp_xml import (
+                        MetadataStream,
+                    )
+                    stm_cls = MetadataStream
+                except ImportError:  # pragma: nocover
+                    pass
+            return stm_cls(data, encoded_data=stream_data)
         else:
             return DictionaryObject(data)
 
 
 class StreamObject(DictionaryObject):
-    """PDF stream object.
+    """
+    PDF stream object.
 
     Essentially, a PDF stream is a dictionary object with a binary blob of
     data attached. This data can be encoded by various filters (not all of which
@@ -1110,30 +1253,34 @@ class StreamObject(DictionaryObject):
     decoding, with :class:`.writer.BasePdfFileWriter` and its subclasses working
     the other way around.
 
-    Note that the :class:`.StreamObject` class manages some of its dictionary
-    keys by itself. This is partly the case for the various ``/Filter``
-    and ``/DecodeParms`` entries, but also for the ``/Length`` entry.
-    The latter will be overwritten as necessary.
+    .. note::
+        The :class:`.StreamObject` class manages some of its dictionary
+        keys by itself. This is partly the case for the various ``/Filter``
+        and ``/DecodeParms`` entries, but also for the ``/Length`` entry.
+        The latter will be overwritten as necessary.
+
+    :param dict_data:
+        The dictionary data for this stream object.
+    :param stream_data:
+        The (unencoded) stream data.
+    :param encoded_data:
+        The encoded stream data.
+
+        .. warning::
+            Ordinarily, a stream can be initialised either from decoded and from
+            encoded data.
+
+            If both `stream_data` and `encoded_data` are provided, the caller
+            is responsible for making sure that both are compatible given the
+            currently relevant filter configuration.
+    :param handler:
+        A reference to the currently active
+        :class:`.pyhanko.pdf_utils.crypt.SecurityHandler`.
+        This is only necessary if the stream requires crypt filters.
     """
 
     def __init__(self, dict_data=None, stream_data=None, encoded_data=None,
                  handler=None):
-        """Initialise a stream with dictionary data and stream data
-        (either encoded or decoded). If both `stream_data` and `encoded_data`
-        are provided, the caller is responsible for making sure that both are
-        compatible given the currently relevant filter configuration.
-
-        :param dict_data:
-            The dictionary data for this stream object.
-        :param stream_data:
-            The (unencoded) stream data.
-        :param encoded_data:
-            The encoded stream data.
-        :param handler:
-            A reference to the currently active
-            :class:`.pyhanko.pdf_utils.crypt.SecurityHandler`.
-            This is only necessary if the stream requires crypt filters.
-        """
         super().__init__(dict_data)
         self._data = stream_data
         self._encoded_data = encoded_data
@@ -1309,11 +1456,12 @@ class StreamObject(DictionaryObject):
             if params:
                 self[pdf_name('/DecodeParms')] = params
         else:
+            # FIXME deal with shortened names for standard filters
             # split cur_filters back into two pieces
             filter_names, param_sets = zip(*cur_filters)
             if not allow_duplicates and filter_name in filter_names:
                 if allow_duplicates is False:
-                    raise ValueError(
+                    raise PdfWriteError(
                         f'Filter {filter_name} has already been applied to '
                         f'this stream.'
                     )
@@ -1446,6 +1594,38 @@ def proxy_encrypted_obj(encrypted_obj, handler):
 
 
 class DecryptedObjectProxy(PdfObject):
+    """
+    Internal proxy class that allows transparent on-demand encryption
+    of objects.
+
+    .. warning::
+        Most public-facing APIs won't leave you to deal with these *directly*
+        (that's half the reason this class exists in the first place), and
+        the API of this class is considered internal.
+
+        However, for reasons related to the historical PyPDF2 codebase from
+        which pyHanko's object handling code ultimately derives, there are
+        some Python builtins that might cause these wrapper objects to
+        inadvertently "leak". Please `tell us about such cases
+        <https://github.com/MatthiasValvekens/pyHanko/discussions>`_ so we can
+        make those types of access more convenient and robust.
+
+    .. danger::
+        The ``__eq__`` implementation on this class is not safe for general use,
+        due to the fact that certain structures in PDF are exempt from
+        encryption. Only compare proxy objects with ``==`` in areas of the
+        document where these exemptions don't apply.
+
+    :param raw_object:
+        A raw object, typically as-parsed from a PDF file.
+    :param handler:
+        The security handler governing this object.
+    """
+
+    raw_object: PdfObject
+    """
+    The underlying raw object, in its encrypted state.
+    """
 
     def __init__(self, raw_object: PdfObject, handler):
         self.raw_object = raw_object
@@ -1453,7 +1633,16 @@ class DecryptedObjectProxy(PdfObject):
         self.handler = handler
 
     @property
-    def decrypted(self):
+    def decrypted(self) -> PdfObject:
+        """
+        The decrypted PDF object exposed as a property.
+
+        If this object is a container object, its constituent parts will be
+        wrapped in :class:`.DecryptedObjectProxy` as well, in order to defer
+        further decryption until the values are requested through a getter
+        method on the container.
+        """
+
         decrypted = self._decrypted
         if decrypted is not None:
             return decrypted
