@@ -5,389 +5,49 @@ Contains code from the PyPDF2 project; see :ref:`here <pypdf2-license>`
 for the original license.
 """
 
-import enum
 import os
-import struct
-from dataclasses import dataclass
-from io import BytesIO
-from typing import Iterable, List, Optional, Tuple, Union
+from typing import List, Optional, Tuple, Union
 
 from asn1crypto import x509
 
-from pyhanko import __version__
 from pyhanko.pdf_utils import content, generic
 from pyhanko.pdf_utils.crypt import (
     PubKeySecurityHandler,
     SecurityHandler,
     StandardSecurityHandler,
 )
-from pyhanko.pdf_utils.generic import pdf_name, pdf_string
+from pyhanko.pdf_utils.extensions import (
+    DeveloperExtension,
+    DevExtensionMultivalued,
+)
+from pyhanko.pdf_utils.generic import pdf_name
+from pyhanko.pdf_utils.metadata.info import (
+    update_info_dict,
+    view_from_info_dict,
+)
+from pyhanko.pdf_utils.metadata.model import DocumentMetadata
 from pyhanko.pdf_utils.misc import (
+    IndirectObjectExpected,
+    PdfError,
     PdfReadError,
     PdfWriteError,
     instance_test,
-    peek,
 )
 from pyhanko.pdf_utils.rw_common import PdfHandler
+from pyhanko.pdf_utils.xref import (
+    OBJSTREAM_FORBIDDEN,
+    ObjectStream,
+    PositionDict,
+    XRefStream,
+    write_xref_table,
+)
 
 __all__ = [
-    'ObjectStream', 'BasePdfFileWriter',
+    'BasePdfFileWriter',
     'PageObject', 'PdfFileWriter', 'init_xobject_dictionary',
-    'copy_into_new_writer', 'DeveloperExtension', 'DevExtensionMultivalued'
+    'copy_into_new_writer'
 ]
 
-VENDOR = 'pyHanko ' + __version__
-
-
-OBJSTREAM_FORBIDDEN = (generic.IndirectObject, generic.StreamObject)
-
-
-class ObjectStream:
-    """
-    Utility class to collect objects into a PDF object stream.
-
-    Object streams are mainly useful for space efficiency reasons.
-    They allow related objects to be grouped & compressed together in a
-    more flexible manner.
-
-
-    .. warning::
-        Object streams can only be used in files with a cross-reference
-        stream, as opposed to a classical XRef table.
-        In particular, this means that incremental updates to files with a
-        legacy XRef table cannot contain object streams either.
-        See § 7.5.7 in ISO 32000-1 for further details.
-
-    .. danger::
-        Use :meth:`.BasePdfFileWriter.prepare_object_stream` to create instances
-        of object streams. The `__init__` function is internal API.
-
-    """
-
-    def __init__(self, writer: 'BasePdfFileWriter', compress=True):
-        self._obj_refs = {}
-        self.compress = compress
-        self.writer = writer
-        self._ref = None
-
-    def add_object(self, idnum: int, obj: generic.PdfObject):
-        """
-        Add an object to an object stream.
-        Note that objects in object streams always have their generation number
-        set to `0` by definition.
-
-        :param idnum:
-            The object's ID number.
-        :param obj:
-            The object to embed into the object stream.
-        :raise TypeError:
-            Raised if ``obj`` is an instance of :class:`~.generic.StreamObject`
-            or :class:`~.generic.IndirectObject`.
-        """
-
-        if isinstance(obj, OBJSTREAM_FORBIDDEN):
-            raise TypeError(
-                'Stream objects and bare references cannot be embedded into '
-                'object streams.'
-            )
-        self._obj_refs[idnum] = obj
-
-    def register_and_emit(self):
-        """
-        Internal method to flush an object stream as part of the file
-        writing process.
-        """
-        stream_ref = self._ref
-        objects = self._obj_refs
-        if objects and stream_ref is None:
-            # first, register the object stream object
-            #  (will get written later)
-            stream_ref = self._ref \
-                = self.writer.add_object(self.as_pdf_object())
-        # loop over all objects in the stream, and prepare
-        # the data to put in the XRef table
-        for ix, (idnum, obj) in enumerate(objects.items()):
-            yield idnum, (stream_ref.idnum, ix)
-
-    def as_pdf_object(self) -> generic.StreamObject:
-        """
-        Render the object stream to a PDF stream object
-
-        :return: An instance of :class:`~.generic.StreamObject`.
-        """
-        stream_header = BytesIO()
-        main_body = BytesIO()
-        for idnum, obj in self._obj_refs.items():
-            offset = main_body.tell()
-            obj.write_to_stream(main_body, None)
-            stream_header.write(b'%d %d ' % (idnum, offset))
-
-        first_obj_offset = stream_header.tell()
-        stream_header.seek(0)
-        sh_bytes = stream_header.read(first_obj_offset)
-        stream_data = sh_bytes + main_body.getvalue()
-        stream_object = generic.StreamObject({
-            pdf_name('/Type'): pdf_name('/ObjStm'),
-            pdf_name('/N'): generic.NumberObject(len(self._obj_refs)),
-            pdf_name('/First'): generic.NumberObject(first_obj_offset)
-        }, stream_data=stream_data)
-        if self.compress:
-            stream_object.compress()
-        return stream_object
-
-
-class DevExtensionMultivalued(enum.Enum):
-    """
-    Setting indicating how an extension is expected to behave well w.r.t.
-    the new mechanism for multivalued extensions in ISO 32000-2:2020.
-    """
-
-    ALWAYS = enum.auto()
-    """
-    Always serialise this extension as a multivalued extension.
-    """
-
-    NEVER = enum.auto()
-    """
-    Never serialise this extension as a multivalued extension.
-    """
-
-    MAYBE = enum.auto()
-    """
-    Make this extension single-valued whenever possible, but allow multiple
-    values as well, e.g. when a different but non-comparable extension with
-    the same prefix is already present in the file.
-    """
-
-
-@dataclass(frozen=True)
-class DeveloperExtension:
-    """
-    PDF developer extension designation.
-    """
-
-    prefix_name: generic.NameObject
-    """
-    Registered developer prefix.
-    """
-
-    base_version: generic.NameObject
-    """
-    Base version on to which the extension applies.
-    """
-
-    extension_level: int
-    """
-    Extension level.
-    """
-
-    url: Optional[str] = None
-    """
-    Optional URL linking to the extension's documentation.
-    """
-
-    extension_revision: Optional[str] = None
-    """
-    Optional extra revision information. Not comparable.
-    """
-
-    compare_by_level: bool = False
-    """
-    Compare developer extensions by level number.
-    If this value is ``True`` and a copy of this extension already exists in the
-    target file with a higher level number, do not override it.
-    If one exists with a lower level number, override it.
-    
-    If this value is ``False``, the decision is based on :attr:`subsumed_by`
-    and :attr:`subsumes`.
-    
-    .. warning::
-        It is generally not safe to assume that extension levels are used as a
-        versioning system (i.e. that higher extension levels supersede lower
-        ones), hence why the default is ``False``.
-    """
-
-    subsumed_by: Iterable[int] = ()
-    """
-    List of extension levels that would subsume this one. If one of these is
-    present in the extensions dictionary, attempting to register this extension
-    will not override it.
-    
-    Default value: empty.
-    
-    .. warning::
-        This parameter is ignored if :class:`compare_by_level` is ``True``.
-    """
-
-    subsumes: Iterable[int] = ()
-    """
-    List of extensions explicitly subsumed by this one. If one of these is
-    present in the extensions dictionary, attempting to register this extension
-    will override it.
-    
-    Default value: empty.
-    
-    .. warning::
-        This parameter is ignored if :class:`compare_by_level` is ``True``.
-    """
-
-    multivalued: DevExtensionMultivalued = DevExtensionMultivalued.MAYBE
-    """
-    Setting indicating whether this extension is expected to behave well w.r.t.
-    the new mechanism for multivalued extensions in ISO 32000-2:2020.
-    """
-
-    def as_pdf_object(self) -> generic.DictionaryObject:
-        """
-        Format the data in this object into a PDF dictionary for registration
-        into the `/Extensions` dictionary.
-
-        :return:
-            A :class:`.generic.DictionaryObject`.
-        """
-
-        result = generic.DictionaryObject({
-            pdf_name('/Type'): pdf_name('/DeveloperExtensions'),
-            pdf_name('/BaseVersion'): self.base_version,
-            pdf_name('/ExtensionLevel'): generic.NumberObject(
-                self.extension_level
-            ),
-        })
-        if self.url is not None:
-            result['/URL'] = generic.TextStringObject(self.url)
-        if self.extension_revision is not None:
-            result['/ExtensionRevision'] = generic.TextStringObject(
-                self.extension_revision
-            )
-        return result
-
-
-def _contiguous_xref_chunks(position_dict):
-    """
-    Helper method to divide the XRef table (or stream) into contiguous chunks.
-    """
-    previous_idnum = None
-    current_chunk = []
-
-    if not position_dict.keys():
-        # return immediately, there are no objects
-        return
-
-    # iterate over keys in object ID order
-    key_iter = sorted(position_dict.keys(), key=lambda t: t[1])
-    (_, first_idnum), key_iter = peek(key_iter)
-    for ix in key_iter:
-        generation, idnum = ix
-
-        # the idnum jumped, so yield the current chunk
-        # and start a new one
-        if current_chunk and idnum != previous_idnum + 1:
-            yield first_idnum, current_chunk
-            current_chunk = []
-            first_idnum = idnum
-
-        # append the object reference to the current chunk
-        # (xref table requires position and generation entries)
-        current_chunk.append((position_dict[ix], generation))
-        previous_idnum = idnum
-
-    # there is always at least one chunk, so this is fine
-    yield first_idnum, current_chunk
-
-
-def _write_xref_table(stream, position_dict):
-    xref_location = stream.tell()
-    stream.write(b'xref\n')
-    # Insert xref table subsections in contiguous chunks.
-    # This is necessarily more complicated than the implementation
-    # in PyPDF2 (see ISO 32000 § 7.5.4, esp. on updates), since
-    # we need to handle incremental updates correctly.
-    subsections = _contiguous_xref_chunks(position_dict)
-
-    def write_header(idnum, length):
-        header = '%d %d\n' % (idnum, length)
-        stream.write(header.encode('ascii'))
-
-    def write_subsection(chunk):
-        for position, generation in chunk:
-            entry = "%010d %05d n \n" % (position, generation)
-            stream.write(entry.encode('ascii'))
-
-    try:
-        first_idnum, subsection = next(subsections)
-    except StopIteration:
-        # no updates, just write '0 0' and be done with it
-        stream.write(b'0 0\n')
-        return xref_location
-    # TODO support deleting objects
-    # case distinction: in contrast with the above we have to ensure that
-    # everything is written in one chunk when *not* doing incremental updates.
-    # In particular, this applies to the null object
-    null_obj_ref = b'0000000000 65535 f \n'
-    if first_idnum == 1:
-        # integrate the null object into the first subsection
-        write_header(0, len(subsection) + 1)
-        stream.write(null_obj_ref)
-        write_subsection(subsection)
-    else:
-        # insert origin of linked list of freed objects, and then the first
-        # subsection, as usual
-        stream.write(b'0 1\n')
-        stream.write(null_obj_ref)
-        write_header(first_idnum, len(subsection))
-        write_subsection(subsection)
-    for first_idnum, subsection in subsections:
-        # subsection header: list first object ID + length of subsection
-        write_header(first_idnum, len(subsection))
-        write_subsection(subsection)
-
-    return xref_location
-
-
-class XRefStream(generic.StreamObject):
-
-    def __init__(self, position_dict):
-        super().__init__()
-        self.position_dict = position_dict
-
-        # type indicator is one byte wide
-        # we use longs to indicate positions of objects (>Q)
-        # two more bytes for the generation number of an uncompressed object
-        widths = map(generic.NumberObject, (1, 8, 2))
-        self.update({
-            pdf_name('/W'): generic.ArrayObject(widths),
-            pdf_name('/Type'): pdf_name('/XRef'),
-        })
-
-    def write_to_stream(self, stream, handler=None, container_ref=None):
-        # the caller is responsible for making sure that the stream
-        # is registered in the position dictionary
-
-        index = [0, 1]
-        subsections = _contiguous_xref_chunks(self.position_dict)
-        stream_content = BytesIO()
-        # write null object
-        stream_content.write(b'\x00' * 9 + b'\xff\xff')
-        for first_idnum, subsection in subsections:
-            index += [first_idnum, len(subsection)]
-            for position, generation in subsection:
-                if isinstance(position, tuple):
-                    # reference to object in object stream
-                    assert generation == 0
-                    obj_stream_num, ix = position
-                    stream_content.write(b'\x02')
-                    stream_content.write(struct.pack('>Q', obj_stream_num))
-                    stream_content.write(struct.pack('>H', ix))
-                else:
-                    stream_content.write(b'\x01')
-                    stream_content.write(struct.pack('>Q', position))
-                    stream_content.write(struct.pack('>H', generation))
-        index_entry = generic.ArrayObject(map(generic.NumberObject, index))
-
-        self[pdf_name('/Index')] = index_entry
-        self._data = stream_content.getbuffer()
-        super().write_to_stream(stream, None)
-
 
 # TODO move this to content.py?
 def init_xobject_dictionary(command_stream: bytes, box_width, box_height,
@@ -455,11 +115,16 @@ class BasePdfFileWriter(PdfHandler):
         self._encrypt = self._encrypt_key = None
         self._document_id = document_id
         self.stream_xrefs = stream_xrefs
-        if info is not None and \
-                not isinstance(info, generic.IndirectObject):
-            self._info = self.add_object(info)
-        else:
-            self._info = info
+        info_ref = None
+        if info is not None:
+            if not isinstance(info, generic.IndirectObject):
+                info_ref = self.add_object(info)
+            else:
+                info_ref = generic.IndirectObject(
+                    info.idnum, info.generation, self
+                )
+        self._info = info_ref
+        self._meta: DocumentMetadata = DocumentMetadata()
 
         self._font_resources = {}
 
@@ -473,6 +138,23 @@ class BasePdfFileWriter(PdfHandler):
 
         return fsc
 
+    @property
+    def document_meta(self) -> DocumentMetadata:
+        return self._meta
+
+    @property
+    def document_meta_view(self) -> DocumentMetadata:
+        # we need the view_over (not just a copy) because
+        # e.g. copy_into_new_writer will populate the info dict with
+        # base values that can then be "cleanly" overridden in the
+        # high-level API without destroying any unsupported entries
+        # in the original info dict
+        if self._info:
+            base = view_from_info_dict(self._info.get_object())
+        else:
+            base = DocumentMetadata()
+        return self._meta.view_over(base)
+
     def ensure_output_version(self, version):
         if self.output_version < version:
             self.output_version = version
@@ -565,7 +247,9 @@ class BasePdfFileWriter(PdfHandler):
 
         # check if the extension is already registered,
         try:
-            cur_ext_value = extensions[ext.prefix_name]
+            cur_ext_value = extensions.raw_get(
+                ext.prefix_name, decrypt=generic.EncryptedObjAccess.RAW
+            )
         except KeyError:
             cur_ext_value = None
         extension_dicts = ()
@@ -658,9 +342,9 @@ class BasePdfFileWriter(PdfHandler):
                 )
         self.update_container(extensions)
 
-    def get_object(self, ido):
+    def get_object(self, ido, as_metadata_stream: bool = False):
         if ido.pdf not in self._resolves_objs_from:
-            raise ValueError(
+            raise PdfError(
                 f'Reference {ido} has no relation to this PDF writer.'
             )
         idnum = ido.idnum
@@ -756,7 +440,7 @@ class BasePdfFileWriter(PdfHandler):
             raise PdfWriteError(
                 'Object streams require Xref streams to be enabled.'
             )
-        stream = ObjectStream(self, compress=compress)
+        stream = ObjectStream(compress=compress)
         self.object_streams.append(stream)
         return stream
 
@@ -766,11 +450,30 @@ class BasePdfFileWriter(PdfHandler):
     def _assign_security_handler(self, sh: SecurityHandler):
         self.security_handler = sh
         self._encrypt = self.add_object(sh.as_pdf_object())
+        min_pdf_version = sh.get_min_pdf_version()
+        if min_pdf_version is not None:
+            self.ensure_output_version(min_pdf_version)
 
-    def _write_objects(self, stream, object_position_dict):
+    def _flush_obj_stream(self, obj_stm: ObjectStream):
+        """
+        Internal method to flush an object stream as part of the file
+        writing process.
+        """
+        stream_ref = obj_stm.ref
+        if obj_stm and not stream_ref:
+            # first, register the object stream object
+            #  (will get written later)
+            obj_stm.ref = stream_ref = self.add_object(obj_stm.as_pdf_object())
+        # loop over all objects in the stream, and prepare
+        # the data to put in the XRef table
+        for ix, (idnum, obj) in enumerate(obj_stm):
+            yield idnum, (stream_ref.idnum, ix)
+
+    def _write_objects(self, stream,
+                       object_position_dict: PositionDict):
         # deal with objects in object streams first
         for obj_stream in self.object_streams:
-            for idnum, pos_record in obj_stream.register_and_emit():
+            for idnum, pos_record in self._flush_obj_stream(obj_stream):
                 object_position_dict[(0, idnum)] = pos_record
 
         for ix in sorted(self.objects.keys()):
@@ -787,6 +490,62 @@ class BasePdfFileWriter(PdfHandler):
             obj.write_to_stream(stream, handler, container_ref)
             stream.write(b'\nendobj\n')
 
+    def _prep_dom_for_writing(self):
+        # ensure that all font resources are flushed
+        for fsc in self._font_resources.values():
+            for engine in fsc.subsets.values():
+                engine.prepare_write()
+
+        self._update_meta()
+
+    def _update_meta(self):
+        try:
+            # delayed import since the namespace registration operation
+            # is global (thank you ElementTree...)
+            # also, we want the XML dep(s) to be optional
+            # noinspection PyUnresolvedReferences
+            from pyhanko.pdf_utils.metadata import xmp_xml
+            need_xmp = (
+                self._meta.xmp_unmanaged
+                or bool(self._meta.xmp_extra)
+                or self.output_version >= (2, 0)
+                or '/Metadata' in self.root
+            )
+        except ImportError:  # pragma: nocover
+            need_xmp = False
+
+        self._meta.last_modified = 'now'
+        if self._info is not None:
+            mod = update_info_dict(
+                self._meta, self._info.get_object(),
+                # if we write XMP, we only update existing entries
+                only_update_existing=need_xmp
+            )
+            if mod:
+                self.mark_update(self._info)
+        elif not need_xmp:
+            info_dict = generic.DictionaryObject()
+            update_info_dict(self._meta, info_dict)
+            self._info = self.add_object(info_dict)
+            # if there's no info dict, and we're going to write XMP anyhow,
+            # don't bother creating one
+
+        if need_xmp:
+            meta_stm = None
+            if '/Metadata' in self.root:
+                meta_obj = self.root['/Metadata']
+                if isinstance(meta_obj, xmp_xml.MetadataStream):
+                    meta_stm = meta_obj
+                    meta_stm.update_xmp_with_meta(self._meta)
+                    self.update_container(meta_stm)
+            if meta_stm is None:
+                meta_stm = xmp_xml.MetadataStream.from_xmp(
+                    xmp_xml.update_xmp_with_meta(self._meta)
+                )
+                self.root['/Metadata'] = self.add_object(meta_stm)
+                self.update_root()
+            self.update_root()
+
     def _populate_trailer(self, trailer):
         # prepare trailer dictionary entries
         trailer[pdf_name('/Root')] = self._root
@@ -811,16 +570,11 @@ class BasePdfFileWriter(PdfHandler):
         :param stream:
             A writable output stream.
         """
+        self._prep_dom_for_writing()
         self._write(stream)
 
     def _write(self, stream, skip_header=False):
-
-        # ensure that all font resources are flushed
-        for fsc in self._font_resources.values():
-            for engine in fsc.subsets.values():
-                engine.prepare_write()
-
-        object_positions = {}
+        object_positions: PositionDict = {}
 
         if self.stream_xrefs:
             trailer = XRefStream(object_positions)
@@ -840,12 +594,12 @@ class BasePdfFileWriter(PdfHandler):
             object_positions[(0, xrefs_id)] = xref_location
             trailer[pdf_name('/Size')] = generic.NumberObject(xrefs_id + 1)
             # write XRef stream
-            stream.write(('%d %d obj' % (xrefs_id, 0)).encode('ascii'))
+            stream.write(('%d %d obj\n' % (xrefs_id, 0)).encode('ascii'))
             trailer.write_to_stream(stream, None)
             stream.write(b'\nendobj\n')
         else:
             # classical xref table
-            xref_location = _write_xref_table(stream, object_positions)
+            xref_location = write_xref_table(stream, object_positions)
             trailer[pdf_name('/Size')] = generic.NumberObject(
                 self._lastobj_id + 1
             )
@@ -898,9 +652,9 @@ class BasePdfFileWriter(PdfHandler):
             A reference to the newly inserted page.
         """
         if new_page['/Type'] != pdf_name('/Page'):
-            raise ValueError('Not a page object')
+            raise PdfWriteError('Not a page object')
         if '/Parent' in new_page:
-            raise ValueError('/Parent must not be set.')
+            raise PdfWriteError('/Parent must not be set.')
 
         page_tree_root_ref = self.root.raw_get('/Pages')
         if after is None:
@@ -918,7 +672,7 @@ class BasePdfFileWriter(PdfHandler):
         try:
             kids = pages_obj['/Kids']
         except KeyError:  # pragma: nocover
-            raise ValueError('/Pages must have /Kids')
+            raise PdfError('/Pages must have /Kids')
 
         # increase page count for all parents
         parent = pages_obj
@@ -1001,15 +755,41 @@ class BasePdfFileWriter(PdfHandler):
             raw_dict = {
                 k: self._import_object(v, reference_map, obj_stream)
                 for k, v in obj.items()
+                if k != '/Metadata'
             }
+            try:
+                # make sure to import metadata streams as such
+                meta_ref = obj.get_value_as_reference('/Metadata')
+                # ensure a MetadataStream object ends up in the cache
+                meta_ref.get_pdf_handler().get_object(
+                    meta_ref, as_metadata_stream=True
+                )
+                # ...then import the reference
+                raw_dict['/Metadata'] = self._import_object(
+                    generic.IndirectObject(
+                        meta_ref.idnum, meta_ref.generation, meta_ref.pdf
+                    ),
+                    reference_map,
+                    obj_stream
+                )
+            except (KeyError, IndirectObjectExpected):
+                pass
+
             if isinstance(obj, generic.StreamObject):
+                stm_cls = generic.StreamObject
+                # again, make sure to import metadata streams as such
+                try:
+                    # noinspection PyUnresolvedReferences
+                    from pyhanko.pdf_utils.metadata import xmp_xml
+                    if isinstance(obj, xmp_xml.MetadataStream):
+                        stm_cls = xmp_xml.MetadataStream
+                except ImportError:  # pragma: nocover
+                    pass
                 # In the vast majority of use cases, I'd expect the content
                 # to be available in encoded form by default.
                 # By initialising the stream object in this way, we avoid
                 # a potentially costly decoding operation.
-                return generic.StreamObject(
-                    raw_dict, encoded_data=obj.encoded_data
-                )
+                return stm_cls(raw_dict, encoded_data=obj.encoded_data)
             else:
                 return generic.DictionaryObject(raw_dict)
         elif isinstance(obj, generic.ArrayObject):
@@ -1149,7 +929,7 @@ class BasePdfFileWriter(PdfHandler):
                 # mark the page to be updated as well
                 self.mark_update(page_obj_ref)
             else:
-                raise ValueError('Unexpected type for page /Contents')
+                raise PdfError('Unexpected type for page /Contents')
         elif isinstance(contents_ref, generic.ArrayObject):
             # make /Contents an indirect array, and append our stream
             contents = contents_ref
@@ -1160,7 +940,7 @@ class BasePdfFileWriter(PdfHandler):
             page_obj[pdf_name('/Contents')] = self.add_object(contents)
             self.mark_update(page_obj_ref)
         else:
-            raise ValueError('Unexpected type for page /Contents')
+            raise PdfError('Unexpected type for page /Contents')
 
         if resources is None:
             return
@@ -1240,7 +1020,7 @@ class BasePdfFileWriter(PdfHandler):
             elif isinstance(orig_value, generic.DictionaryObject):
                 for key_, value_ in value.items():
                     if key_ in orig_value:
-                        raise ValueError(
+                        raise PdfError(
                             'Naming conflict in resource of type %s: '
                             'key %s occurs in both.' % (key, key_)
                         )
@@ -1296,12 +1076,6 @@ class PdfFileWriter(BasePdfFileWriter):
         id2 = generic.ByteStringObject(os.urandom(16))
         id_obj = generic.ArrayObject([id1, id2])
 
-        # info object
-        if info is None:
-            info = generic.DictionaryObject({
-                pdf_name('/Producer'): pdf_string(VENDOR)
-            })
-
         self._custom_trailer_entries = {}
         super().__init__(root, info, id_obj, stream_xrefs=stream_xrefs)
 
@@ -1321,7 +1095,7 @@ class PdfFileWriter(BasePdfFileWriter):
         # as binary (see § 7.5.2 in ISO 32000-1)
         stream.write(b'%\xc2\xa5\xc2\xb1\xc3\xab\n')
 
-    def encrypt(self, owner_pass, user_pass=None):
+    def encrypt(self, owner_pass, user_pass=None, **kwargs):
         """
         Mark this document to be encrypted with PDF 2.0 encryption (AES-256).
 
@@ -1355,12 +1129,16 @@ class PdfFileWriter(BasePdfFileWriter):
         :param user_pass:
             The desired user password (defaults to the owner password
             if not specified)
+        :param kwargs:
+            Other keyword arguments to be passed to
+            :meth:`.StandardSecurityHandler.build_from_pw`.
         """
-        self.output_version = (2, 0)
-        sh = StandardSecurityHandler.build_from_pw(owner_pass, user_pass)
+        sh = StandardSecurityHandler.build_from_pw(
+            owner_pass, user_pass, **kwargs
+        )
         self._assign_security_handler(sh)
 
-    def encrypt_pubkey(self, recipients: List[x509.Certificate]):
+    def encrypt_pubkey(self, recipients: List[x509.Certificate], **kwargs):
         """
         Mark this document to be encrypted with PDF 2.0 public key encryption.
         The certificates passed in should be RSA certificates.
@@ -1377,9 +1155,12 @@ class PdfFileWriter(BasePdfFileWriter):
         :param recipients:
             Certificates of the recipients that should be able to decrypt
             the document.
+        :param kwargs:
+            Other keyword arguments to be passed to
+            :meth:`.PubKeySecurityHandler.build_from_certs`.
         """
         self.output_version = (2, 0)
-        sh = PubKeySecurityHandler.build_from_certs(recipients)
+        sh = PubKeySecurityHandler.build_from_certs(recipients, **kwargs)
         self._assign_security_handler(sh)
 
     def set_custom_trailer_entry(self, key: generic.NameObject,
@@ -1406,22 +1187,8 @@ class PdfFileWriter(BasePdfFileWriter):
         super()._populate_trailer(trailer)
 
 
-# skip Trapped because we don't care, and because it's not a string
-PERMISSIBLE_INFO_KEYS = {
-    '/Title', '/Author', '/Subject', '/Keywords', '/Creator', '/Producer',
-    '/CreationDate', '/ModDate'
-}
-
-
-def _clone_info_dict(input_info_dict):
-    return generic.DictionaryObject({
-        k: pdf_string(str(v.get_object()))
-        for k, v in input_info_dict.items()
-        if k in PERMISSIBLE_INFO_KEYS
-    })
-
-
-def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
+def copy_into_new_writer(input_handler: PdfHandler,
+                         writer_kwargs: dict = None) -> PdfFileWriter:
     """
     Copy all objects in a given PDF handler into a new :class:`.PdfFileWriter`.
     This operation will attempt to preserve the document catalog
@@ -1437,30 +1204,18 @@ def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
 
     :param input_handler:
         :class:`.PdfHandler` to source objects from.
+    :param writer_kwargs:
+        Keyword arguments to pass to the writer.
     :return:
         New :class:`.PdfFileWriter` containing all objects from the input
         handler.
     """
-    try:
-        info_ref = input_handler.trailer_view.raw_get('/Info')
-    except KeyError:
-        info_ref = None
 
-    if info_ref is not None:
-        info = _clone_info_dict(info_ref.get_object())
-        try:
-            producer_string = info['/Producer']
-            if VENDOR not in producer_string:
-                producer_string = pdf_string(f"{producer_string}; {VENDOR}")
-        except KeyError:
-            producer_string = pdf_string(VENDOR)
-        # always override this
-        info['/Producer'] = producer_string
-    else:
-        info = None
+    writer_kwargs = writer_kwargs or {}
+    writer_kwargs.setdefault("stream_xrefs", False)
 
     # TODO try to be more clever with object streams
-    w = PdfFileWriter(init_page_tree=False, stream_xrefs=False, info=info)
+    w = PdfFileWriter(init_page_tree=False, **writer_kwargs)
     input_root_ref = input_handler.root_ref
     output_root_ref = w.root_ref
     # call _import_object in such a way that we translate the input handler's
@@ -1476,4 +1231,19 @@ def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
     ix = (output_root_ref.generation, output_root_ref.idnum)
     w.objects[ix] = new_root_dict
 
+    if "info" not in writer_kwargs:
+        try:
+            # migrate the info dict. We do this low-level to avoid issues
+            # with the producer string handling, and to keep a nice separation
+            # between user-supplied metadata values and values that were present
+            # in the original doc.
+            info_dict = input_handler.trailer_view['/Info']
+        except KeyError:
+            info_dict = None
+        if info_dict is not None:
+            imported_info = w._import_object(
+                info_dict, reference_map={}, obj_stream=None
+            )
+            w._info = w.add_object(imported_info)
+
     return w
