@@ -9,7 +9,7 @@ CMS is defined in :rfc:`5652`. To parse CMS messages, pyHanko relies heavily on
 import hashlib
 import logging
 from dataclasses import dataclass
-from typing import List, Optional, Union
+from typing import IO, Iterable, List, Optional, Tuple, Union
 
 from asn1crypto import algos, cms, keys, pem, tsp, x509
 from cryptography.hazmat.primitives import hashes, serialization
@@ -21,6 +21,8 @@ from pyhanko_certvalidator.registry import (
     SimpleCertificateStore,
 )
 
+from pyhanko.pdf_utils import misc
+
 __all__ = [
     'simple_cms_attribute',
     'find_cms_attribute', 'find_unique_cms_attribute',
@@ -32,12 +34,31 @@ __all__ = [
     'load_private_key_from_pemder', 'get_pyca_cryptography_hash',
     'optimal_pss_params', 'process_pss_params', 'as_signing_certificate',
     'as_signing_certificate_v2', 'match_issuer_serial',
-    'check_ess_certid',
+    'check_ess_certid', 'CMSExtractionError', 'byte_range_digest',
+    'ValueErrorWithMessage', 'CMSStructuralError'
 ]
 
 logger = logging.getLogger(__name__)
 
 
+class ValueErrorWithMessage(ValueError):
+    """
+    Value error with a failure message attribute that can be conveniently
+    extracted, instead of having to rely on extracting exception args
+    generically.
+    """
+    def __init__(self, failure_message):
+        self.failure_message = str(failure_message)
+
+
+class CMSStructuralError(ValueErrorWithMessage):
+    """Structural error in a CMS object."""
+
+
+class CMSExtractionError(ValueErrorWithMessage):
+    pass
+
+
 def simple_cms_attribute(attr_type, value):
     """
     Convenience method to quickly construct a CMS attribute object with
@@ -68,6 +89,11 @@ def find_cms_attribute(attrs, name):
     """
     Find and return CMS attribute values of a given type.
 
+    .. note::
+        This function will also check for duplicates, but not in the sense
+        of multivalued attributes. In other words: multivalued attributes
+        are allowed; listing the same attribute OID more than once is not.
+
     :param attrs:
         The :class:`.cms.CMSAttributes` object.
     :param name:
@@ -77,12 +103,24 @@ def find_cms_attribute(attrs, name):
     :raise NonexistentAttributeError:
         Raised when no such type entry could be found in the
         :class:`.cms.CMSAttributes` object.
+    :raise CMSStructuralError:
+        Raised if the given OID occurs more than once.
     """
+
+    found_values = None
     if attrs:
         for attr in attrs:
             if attr['type'].native == name:
-                return attr['values']
-    raise NonexistentAttributeError(f'Unable to locate attribute {name}.')
+                if found_values is not None:
+                    raise CMSStructuralError(
+                        f"Attribute {name!r} was duplicated"
+                    )
+                found_values = attr['values']
+
+    if found_values is not None:
+        return found_values
+    else:
+        raise NonexistentAttributeError(f'Unable to locate attribute {name}.')
 
 
 def find_unique_cms_attribute(attrs, name):
@@ -236,9 +274,22 @@ def match_issuer_serial(expected_issuer_serial:
                 expected_issuer[0].name != 'directory_name':
             return False
         expected_issuer = expected_issuer[0].chosen
+
+    # Some certificates contain name elements that can't be compared, e.g.
+    # because they violate the stringprep profile for directory strings.
+    # Instead of failing those outright, we attempt a byte-for-byte comparison
+    # first. If the encoding happens to match exactly, we're good. If not,
+    # we invoke (asn1crypto's implementation of) the directory name matching
+    # routine.
+    try:
+        issuer_match = (
+            expected_issuer.dump() == cert.issuer.dump()
+            or expected_issuer == cert.issuer
+        )
+    except ValueError:
+        issuer_match = False
     return (
-        expected_issuer == cert.issuer
-        and expected_issuer_serial['serial_number'] == serial_asn1
+        issuer_match and expected_issuer_serial['serial_number'] == serial_asn1
     )
 
 
@@ -246,7 +297,9 @@ class SigningError(ValueError):
     """
     Error encountered while signing a file.
     """
-    pass
+    def __init__(self, msg: str, *args):
+        self.msg = msg
+        super().__init__(msg, *args)
 
 
 class UnacceptableSignerError(SigningError):
@@ -486,7 +539,7 @@ def _partition_certs(certs, signer_info):
         else:
             other_certs.append(c)
     if cert is None:
-        raise ValueError('signer certificate not included in signature')
+        raise CMSExtractionError('signer certificate not included in signature')
     return cert, other_certs
 
 
@@ -505,8 +558,8 @@ def extract_signer_info(signed_data: cms.SignedData) -> cms.SignerInfo:
     try:
         signer_info, = signed_data['signer_infos']
         return signer_info
-    except ValueError:  # pragma: nocover
-        raise ValueError(
+    except ValueError:
+        raise CMSExtractionError(
             'signer_infos should contain exactly one entry'
         )
 
@@ -537,3 +590,38 @@ def extract_certificate_info(signed_data: cms.SignedData) -> SignedDataCerts:
         attribute_certs=attr_certs
     )
     return cert_info
+
+
+def byte_range_digest(stream: IO, byte_range: Iterable[int],
+                      md_algorithm: str,
+                      chunk_size=misc.DEFAULT_CHUNK_SIZE) -> Tuple[int, bytes]:
+    """
+    Internal API to compute byte range digests. Potentially dangerous if used
+    without due caution.
+
+    :param stream:
+        Stream over which to compute the digest. Must support seeking and
+        reading.
+    :param byte_range:
+        The byte range, as a list of (offset, length) pairs, flattened.
+    :param md_algorithm:
+        The message digest algorithm to use.
+    :param chunk_size:
+        The I/O chunk size to use.
+    :return:
+        A tuple of the total digested length, and the actual digest.
+    """
+    md_spec = get_pyca_cryptography_hash(md_algorithm)
+    md = hashes.Hash(md_spec)
+
+    # compute the digest
+    # here, we allow arbitrary byte ranges
+    # for the coverage check, we'll impose more constraints
+    total_len = 0
+    chunk_buf = bytearray(chunk_size)
+    for lo, chunk_len in misc.pair_iter(byte_range):
+        stream.seek(lo)
+        misc.chunked_digest(chunk_buf, stream, md, max_read=chunk_len)
+        total_len += chunk_len
+
+    return total_len, md.finalize()
