@@ -6,13 +6,19 @@ seamlessly plugged into a :class:`~.signers.PdfSigner`.
 import asyncio
 import binascii
 import logging
-from typing import Optional, Set
+import warnings
+from dataclasses import dataclass
+from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
 
-from asn1crypto import x509
+from asn1crypto import algos, core, x509
 from asn1crypto.algos import RSASSAPSSParams
 from cryptography.hazmat.primitives import hashes
 
-from pyhanko.config import PKCS11SignatureConfig
+from pyhanko.config import (
+    PKCS11PinEntryMode,
+    PKCS11SignatureConfig,
+    TokenCriteria,
+)
 from pyhanko.sign.general import (
     CertificateStore,
     SigningError,
@@ -22,8 +28,17 @@ from pyhanko.sign.general import (
 from pyhanko.sign.signers import Signer
 
 try:
-    from pkcs11 import Attribute, ObjectClass, PKCS11Error, Session
-    from pkcs11 import lib as pkcs11_lib
+    from pkcs11 import (
+        MGF,
+        PROTECTED_AUTH,
+        Attribute,
+        Mechanism,
+        ObjectClass,
+        PKCS11Error,
+        Session,
+    )
+    from pkcs11 import lib as p11_lib
+    from pkcs11 import types as p11_types
 except ImportError as e:  # pragma: nocover
     raise ImportError(
         "pyhanko.sign.pkcs11 requires pyHanko to be installed with "
@@ -33,14 +48,260 @@ except ImportError as e:  # pragma: nocover
 
 
 __all__ = [
-    'PKCS11Signer', 'open_pkcs11_session', 'PKCS11SigningContext'
+    'PKCS11Signer', 'open_pkcs11_session', 'PKCS11SigningContext', 'find_token',
+    'select_pkcs11_signing_params'
 ]
 
 logger = logging.getLogger(__name__)
 
 
-def open_pkcs11_session(lib_location, slot_no=None, token_label=None,
-                        user_pin=None) -> Session:
+def criteria_mismatches(criteria: TokenCriteria, token: p11_types.Token) \
+        -> List[Tuple[str, str]]:
+    if criteria is None:
+        return []
+
+    err_items = []
+
+    if criteria.label is not None and token.label != criteria.label:
+        err_items.append(
+            ('label', criteria.label)
+        )
+    if criteria.serial is not None and token.serial != criteria.serial:
+        err_items.append(
+            ('serial', criteria.serial.hex())
+        )
+    return err_items
+
+
+def criteria_satisfied_by(criteria: TokenCriteria, token: p11_types.Token) \
+        -> bool:
+    return not criteria_mismatches(criteria, token)
+
+
+def find_token(slots: List[p11_types.Slot], slot_no: Optional[int] = None,
+               token_criteria: Optional[TokenCriteria] = None) \
+        -> Optional[p11_types.Token]:
+    """
+    Internal helper method to find a token.
+
+    :param slots:
+        The list of slots.
+    :param slot_no:
+        Slot number to use. If not specified, the first slot containing a token
+        satisfying the criteria will be used
+    :param token_criteria:
+        Criteria the token must satisfy.
+    :return:
+        A PKCS#11 token object, or ``None`` if none was found.
+    """
+
+    if token_criteria is None and slot_no is None:
+        if len(slots) == 1:
+            return slots[0].get_token()
+        else:
+            raise PKCS11Error(
+                "Module has more than 1 slot; slot index or token criteria "
+                "must be provided"
+            )
+
+    if slot_no is None:
+        for slot in slots:
+            try:
+                token = slot.get_token()
+                if criteria_satisfied_by(token_criteria, token):
+                    return token
+            except PKCS11Error:
+                continue
+    else:
+        if slot_no >= len(slots):
+            raise PKCS11Error(
+                f"Slot index {slot_no} too large; there are only {len(slots)}"
+            )
+        token = slots[slot_no].get_token()
+        errors = criteria_mismatches(token_criteria, token)
+        if errors:
+            err_str = ", ".join(
+                f"{field} is not {val!r}"
+                for field, val in errors
+            )
+            raise PKCS11Error(
+                f"Token in slot {slot_no} does not satisfy criteria; {err_str}."
+            )
+        return token
+    return None
+
+
+@dataclass(frozen=True)
+class PKCS11SignatureOperationSpec:
+    """
+    Internal helper class to describe how to invoke a signature operation on
+    a key in a PKCS #11 token.
+    """
+
+    sign_kwargs: Dict[str, Any]
+    """
+    Keyword arguments to the ``sign`` function on the key handle.
+    """
+
+    pre_sign_transform: Optional[Callable[[bytes], bytes]]
+    """
+    An optional transformation to apply to the data prior to signing.
+    """
+
+    post_sign_transform: Optional[Callable[[bytes], bytes]]
+    """
+    An optional transformation to apply to the data after signing.
+    """
+
+
+RSA_MECH_MAP = {
+    'sha1': Mechanism.SHA1_RSA_PKCS,
+    'sha224': Mechanism.SHA224_RSA_PKCS,
+    'sha256': Mechanism.SHA256_RSA_PKCS,
+    'sha384': Mechanism.SHA384_RSA_PKCS,
+    'sha512': Mechanism.SHA512_RSA_PKCS,
+}
+
+
+RSASSA_PSS_MECH_MAP = {
+    'sha1': Mechanism.SHA1_RSA_PKCS_PSS,
+    'sha224': Mechanism.SHA224_RSA_PKCS_PSS,
+    'sha256': Mechanism.SHA256_RSA_PKCS_PSS,
+    'sha384': Mechanism.SHA384_RSA_PKCS_PSS,
+    'sha512': Mechanism.SHA512_RSA_PKCS_PSS,
+}
+
+MGF_MECH_MAP = {
+    'sha1': MGF.SHA1,
+    'sha224': MGF.SHA224,
+    'sha256': MGF.SHA256,
+    'sha384': MGF.SHA384,
+    'sha512': MGF.SHA512
+}
+
+
+ECDSA_MECH_MAP = {
+    'sha1': Mechanism.ECDSA_SHA1,
+    'sha224': Mechanism.ECDSA_SHA224,
+    'sha256': Mechanism.ECDSA_SHA256,
+    'sha384': Mechanism.ECDSA_SHA384,
+    'sha512': Mechanism.ECDSA_SHA512,
+}
+
+
+DSA_MECH_MAP = {
+    'sha1': Mechanism.DSA_SHA1,
+    'sha224': Mechanism.DSA_SHA224,
+    'sha256': Mechanism.DSA_SHA256,
+    # These can't be used in CMS IIRC (since the key sizes required
+    # to meaningfully use them are ridiculous),
+    # but they're in the PKCS#11 spec, so let's add them for
+    # completeness
+    'sha384': Mechanism.DSA_SHA384,
+    'sha512': Mechanism.DSA_SHA512,
+}
+
+
+DIGEST_MECH_MAP = {
+    'sha1': Mechanism.SHA_1,
+    'sha224': Mechanism.SHA224,
+    'sha256': Mechanism.SHA256,
+    'sha384': Mechanism.SHA384,
+    'sha512': Mechanism.SHA512,
+}
+
+
+def select_pkcs11_signing_params(
+        signature_mechanism: algos.SignedDigestAlgorithm,
+        digest_algorithm: str,
+        use_raw_mechanism: bool) -> PKCS11SignatureOperationSpec:
+    """
+    Internal helper function to set up a PKCS #11 signing operation.
+
+    :param signature_mechanism:
+        The signature mechanism to use (as an ASN.1 value)
+    :param digest_algorithm:
+        The digest algorithm to use
+    :param use_raw_mechanism:
+        Whether to attempt to use the raw mechanism on pre-hashed data.
+    :return:
+    """
+    from pkcs11.util.dsa import encode_dsa_signature
+    from pkcs11.util.ec import encode_ecdsa_signature
+
+    pre_sign_transform = None
+    post_sign_transform = None
+    kwargs = {}
+
+    try:
+        signature_algo = signature_mechanism.signature_algo
+    except ValueError:
+        signature_algo = signature_mechanism['algorithm'].native
+
+    if signature_algo == 'rsassa_pkcs1v15':
+        if use_raw_mechanism:
+            kwargs['mechanism'] = Mechanism.RSA_PKCS
+            pre_sign_transform = _hash_fully(
+                digest_algorithm, wrap_digest_info=True
+            )
+        else:
+            kwargs['mechanism'] = RSA_MECH_MAP[digest_algorithm]
+    elif signature_algo == 'dsa':
+        if use_raw_mechanism:
+            kwargs['mechanism'] = Mechanism.DSA
+            pre_sign_transform = _hash_fully(
+                digest_algorithm, wrap_digest_info=False
+            )
+        else:
+            kwargs['mechanism'] = DSA_MECH_MAP[digest_algorithm]
+        post_sign_transform = encode_dsa_signature
+    elif signature_algo == 'ecdsa':
+        if use_raw_mechanism:
+            kwargs['mechanism'] = Mechanism.ECDSA
+            pre_sign_transform = _hash_fully(
+                digest_algorithm, wrap_digest_info=False
+            )
+        else:
+            # TODO test these (unsupported in SoftHSMv2 right now)
+            kwargs['mechanism'] = ECDSA_MECH_MAP[digest_algorithm]
+        post_sign_transform = encode_ecdsa_signature
+    elif signature_algo == 'rsassa_pss':
+        if use_raw_mechanism:
+            raise NotImplementedError(
+                "RSASSA-PSS not available in raw mode"
+            )
+        params: RSASSAPSSParams = signature_mechanism['parameters']
+        assert digest_algorithm == \
+               params['hash_algorithm']['algorithm'].native
+
+        # unpack PSS parameters into PKCS#11 language
+        kwargs['mechanism'] = RSASSA_PSS_MECH_MAP[digest_algorithm]
+
+        pss_digest_param = DIGEST_MECH_MAP[digest_algorithm]
+
+        mgf_val = params['mask_gen_algorithm']['parameters']['algorithm'].native
+        pss_mgf_param = MGF_MECH_MAP[mgf_val]
+        pss_salt_len = params['salt_length'].native
+
+        kwargs['mechanism_param'] = (
+            pss_digest_param, pss_mgf_param, pss_salt_len
+        )
+    else:
+        raise NotImplementedError(
+            f"Signature algorithm '{signature_algo}' is not supported."
+        )
+
+    return PKCS11SignatureOperationSpec(
+        sign_kwargs=kwargs,
+        pre_sign_transform=pre_sign_transform,
+        post_sign_transform=post_sign_transform
+    )
+
+
+def open_pkcs11_session(lib_location: str, slot_no: Optional[int] = None,
+                        token_label: Optional[str] = None,
+                        token_criteria: Optional[TokenCriteria] = None,
+                        user_pin: Union[str, object, None] = None) -> Session:
     """
     Open a PKCS#11 session
 
@@ -50,37 +311,40 @@ def open_pkcs11_session(lib_location, slot_no=None, token_label=None,
         Slot number to use. If not specified, the first slot containing a token
         labelled ``token_label`` will be used.
     :param token_label:
+        .. deprecated:: 0.14.0
+            Use ``token_criteria`` instead.
+
         Label of the token to use. If ``None``, there is no constraint.
+    :param token_criteria:
+        Criteria that the token should match.
     :param user_pin:
-        User PIN to use.
+        User PIN to use, or :attr:`.PROTECTED_AUTH`. If ``None``, authentication
+        is skipped.
 
         .. note::
             Some PKCS#11 implementations do not require PIN when the token
             is opened, but will prompt for it out-of-band when signing.
+            Whether :attr:`.PROTECTED_AUTH` or ``None`` is used in this case
+            depends on the implementation.
     :return:
         An open PKCS#11 session object.
     """
-    lib = pkcs11_lib(lib_location)
+    lib = p11_lib(lib_location)
+
+    if token_criteria is None and token_label is not None:
+        warnings.warn(
+            "'token_label' is deprecated, use 'token_criteria' instead",
+            DeprecationWarning
+        )
+        token_criteria = TokenCriteria(label=token_label)
 
     slots = lib.get_slots()
-    token = None
-    if slot_no is None:
-        for slot in slots:
-            try:
-                token = slot.get_token()
-                if token_label is None or token.label == token_label:
-                    break
-            except PKCS11Error:
-                continue
-        if token is None:
-            raise PKCS11Error(
-                f'No token with label {token_label} found'
-                if token_label is not None else 'No token found'
-            )
-    else:
-        token = slots[slot_no].get_token()
-        if token_label is not None and token.label != token_label:
-            raise PKCS11Error(f'Token in slot {slot_no} is not {token_label}.')
+    token = find_token(slots, slot_no=slot_no, token_criteria=token_criteria)
+    if token is None:
+        raise PKCS11Error(
+            f'No token matching criteria {token_criteria!r} found'
+            if token_criteria is not None else 'No token found'
+        )
 
     kwargs = {}
     if user_pin is not None:
@@ -89,6 +353,26 @@ def open_pkcs11_session(lib_location, slot_no=None, token_label=None,
     return token.open(**kwargs)
 
 
+def _format_pull_err_msg(
+        no_results: bool,
+        label: Optional[str] = None,
+        cert_id: Optional[bytes] = None):
+
+    info_strs = []
+    if label is not None:
+        info_strs.append(f"label '{label}'")
+    if cert_id is not None:
+        info_strs.append(
+            f"ID '{binascii.hexlify(cert_id).decode('ascii')}'"
+        )
+    qualifier = f" with {', '.join(info_strs)}" if info_strs else ""
+    if no_results:
+        err = f"Could not find cert{qualifier}."
+    else:
+        err = f"Found more than one cert{qualifier}."
+    return err
+
+
 def _pull_cert(pkcs11_session: Session, label: Optional[str] = None,
                cert_id: Optional[bytes] = None):
 
@@ -108,28 +392,29 @@ def _pull_cert(pkcs11_session: Session, label: Optional[str] = None,
         cert_obj = results[0]
         return x509.Certificate.load(cert_obj[Attribute.VALUE])
     else:
-        info_strs = []
-        if label is not None:
-            info_strs.append(f"label '{label}'")
-        if cert_id is not None:
-            info_strs.append(
-                f"ID '{binascii.hexlify(cert_id).decode('ascii')}'"
-            )
-        qualifier = f" with {', '.join(info_strs)}" if info_strs else ""
-        if not results:
-            err = f"Could not find cert{qualifier}."
-        else:
-            err = f"Found more than one cert{qualifier}."
+        err = _format_pull_err_msg(
+            no_results=not results, label=label, cert_id=cert_id
+        )
         raise PKCS11Error(err)
 
 
-def _hash_fully(digest_algorithm):
+def _hash_fully(digest_algorithm: str, *, wrap_digest_info: bool):
     md_spec = get_pyca_cryptography_hash(digest_algorithm)
 
-    def _h(data: bytes):
+    def _h(data: bytes) -> bytes:
         h = hashes.Hash(md_spec)
         h.update(data)
-        return h.finalize()
+        digest = h.finalize()
+        if wrap_digest_info:
+            return algos.DigestInfo({
+                'digest_algorithm': {
+                    'algorithm': digest_algorithm.lower(),
+                    'parameters': core.Null()
+                },
+                'digest': digest
+            }).dump()
+        else:
+            return digest
 
     return _h
 
@@ -250,6 +535,15 @@ class PKCS11Signer(Signer):
         self._load_objects()
         return self._signing_cert
 
+    def _select_pkcs11_signing_params(self, digest_algorithm: str) \
+            -> PKCS11SignatureOperationSpec:
+        digest_algorithm = digest_algorithm.lower()
+        return select_pkcs11_signing_params(
+            self.get_signature_mechanism(digest_algorithm),
+            digest_algorithm,
+            use_raw_mechanism=self.use_raw_mechanism
+        )
+
     async def async_sign_raw(self, data: bytes,
                              digest_algorithm: str, dry_run=False) -> bytes:
         if dry_run:
@@ -257,109 +551,18 @@ class PKCS11Signer(Signer):
             return b'0' * 512
 
         await self.ensure_objects_loaded()
-        from pkcs11 import MGF, Mechanism, SignMixin
+        from pkcs11 import SignMixin
 
         kh: SignMixin = self._key_handle
-        kwargs = {}
-        digest_algorithm = digest_algorithm.lower()
-        signature_mechanism = self.get_signature_mechanism(digest_algorithm)
-        signature_algo = signature_mechanism.signature_algo
-        pre_sign_transform = None
-        post_sign_transform = None
-        if signature_algo == 'rsassa_pkcs1v15':
-            if self.use_raw_mechanism:
-                raise NotImplementedError(
-                    "RSASSA-PKCS1v15 not available in raw mode"
-                )
-            kwargs['mechanism'] = {
-                'sha1': Mechanism.SHA1_RSA_PKCS,
-                'sha224': Mechanism.SHA224_RSA_PKCS,
-                'sha256': Mechanism.SHA256_RSA_PKCS,
-                'sha384': Mechanism.SHA384_RSA_PKCS,
-                'sha512': Mechanism.SHA512_RSA_PKCS,
-            }[digest_algorithm]
-        elif signature_algo == 'dsa':
-            if self.use_raw_mechanism:
-                raise NotImplementedError("DSA not available in raw mode")
-            kwargs['mechanism'] = {
-                'sha1': Mechanism.DSA_SHA1,
-                'sha224': Mechanism.DSA_SHA224,
-                'sha256': Mechanism.DSA_SHA256,
-                # These can't be used in CMS IIRC (since the key sizes required
-                # to meaningfully use them are ridiculous),
-                # but they're in the PKCS#11 spec, so let's add them for
-                # completeness
-                'sha384': Mechanism.DSA_SHA384,
-                'sha512': Mechanism.DSA_SHA512,
-            }[digest_algorithm]
-            from pkcs11.util.dsa import encode_dsa_signature
-            post_sign_transform = encode_dsa_signature
-        elif signature_algo == 'ecdsa':
-            if self.use_raw_mechanism:
-                kwargs['mechanism'] = Mechanism.ECDSA
-                pre_sign_transform = _hash_fully(digest_algorithm)
-            else:
-                # TODO test these (unsupported in SoftHSMv2 right now)
-                kwargs['mechanism'] = {
-                    'sha1': Mechanism.ECDSA_SHA1,
-                    'sha224': Mechanism.ECDSA_SHA224,
-                    'sha256': Mechanism.ECDSA_SHA256,
-                    'sha384': Mechanism.ECDSA_SHA384,
-                    'sha512': Mechanism.ECDSA_SHA512,
-                }[digest_algorithm]
+        spec = self._select_pkcs11_signing_params(digest_algorithm)
 
-            from pkcs11.util.ec import encode_ecdsa_signature
-            post_sign_transform = encode_ecdsa_signature
-        elif signature_algo == 'rsassa_pss':
-            if self.use_raw_mechanism:
-                raise NotImplementedError(
-                    "RSASSA-PSS not available in raw mode"
-                )
-            params: RSASSAPSSParams = signature_mechanism['parameters']
-            assert digest_algorithm == \
-                   params['hash_algorithm']['algorithm'].native
-
-            # unpack PSS parameters into PKCS#11 language
-            kwargs['mechanism'] = {
-                'sha1': Mechanism.SHA1_RSA_PKCS_PSS,
-                'sha224': Mechanism.SHA224_RSA_PKCS_PSS,
-                'sha256': Mechanism.SHA256_RSA_PKCS_PSS,
-                'sha384': Mechanism.SHA384_RSA_PKCS_PSS,
-                'sha512': Mechanism.SHA512_RSA_PKCS_PSS,
-            }[digest_algorithm]
-
-            pss_digest_param = {
-                'sha1': Mechanism.SHA_1,
-                'sha224': Mechanism.SHA224,
-                'sha256': Mechanism.SHA256,
-                'sha384': Mechanism.SHA384,
-                'sha512': Mechanism.SHA512,
-            }[digest_algorithm]
-
-            pss_mgf_param = {
-                'sha1': MGF.SHA1,
-                'sha224': MGF.SHA224,
-                'sha256': MGF.SHA256,
-                'sha384': MGF.SHA384,
-                'sha512': MGF.SHA512
-            }[params['mask_gen_algorithm']['parameters']['algorithm'].native]
-            pss_salt_len = params['salt_length'].native
-
-            kwargs['mechanism_param'] = (
-                pss_digest_param, pss_mgf_param, pss_salt_len
-            )
-        else:
-            raise PKCS11Error(
-                f"Signature algorithm '{signature_algo}' is not supported."
-            )
-
-        if pre_sign_transform is not None:
-            data = pre_sign_transform(data)
+        if spec.pre_sign_transform is not None:
+            data = spec.pre_sign_transform(data)
 
         def _perform_signature():
-            signature = kh.sign(data, **kwargs)
-            if post_sign_transform is not None:
-                signature = post_sign_transform(signature)
+            signature = kh.sign(data, **spec.sign_kwargs)
+            if spec.post_sign_transform is not None:
+                signature = spec.post_sign_transform(signature)
             return signature
 
         loop = asyncio.get_running_loop()
@@ -429,7 +632,7 @@ class PKCS11Signer(Signer):
             loop = asyncio.get_running_loop()
             await loop.run_in_executor(None, self._load_objects)
             event.set()
-        else:
+        else:  # pragma: nocover
             # some other coroutine is dealing with fetching already,
             # just wait for that one to finish
             await self.__loading_event.wait()
@@ -462,14 +665,24 @@ class PKCS11SigningContext:
         self._session = None
         self._user_pin = user_pin
 
+    def _handle_pin(self):
+        pin = self._user_pin or self.config.user_pin
+        # mode 'PROMPT' is irrelevant for library usage.
+        if pin is not None:
+            pin = str(pin)
+        elif self.config.prompt_pin == PKCS11PinEntryMode.DEFER:
+            pin = PROTECTED_AUTH
+        else:
+            pin = None
+        return pin
+
     def _instantiate(self) -> PKCS11Signer:
         config = self.config
-        pin = self._user_pin or config.user_pin
-        pin = str(pin) if pin is not None else None
+        pin = self._handle_pin()
 
         self._session = session = open_pkcs11_session(
             config.module_path, slot_no=config.slot_no,
-            token_label=config.token_label,
+            token_criteria=config.token_criteria,
             user_pin=pin
         )
         return PKCS11Signer(
