@@ -15,6 +15,7 @@ from pyhanko.sign.fields import (
     FieldMDPSpec,
     MDPPerm,
     SigFieldSpec,
+    apply_sig_field_spec_properties,
     ensure_sig_flags,
     enumerate_sig_fields,
     prepare_sig_field,
@@ -30,12 +31,11 @@ __all__ = [
 ]
 
 
-def docmdp_reference_dictionary(md_algorithm, permission_level: MDPPerm):
+def docmdp_reference_dictionary(permission_level: MDPPerm):
     # this is part of the /Reference entry of the signature object.
     return generic.DictionaryObject({
         pdf_name('/Type'): pdf_name('/SigRef'),
         pdf_name('/TransformMethod'): pdf_name('/DocMDP'),
-        pdf_name('/DigestMethod'): pdf_name('/' + md_algorithm.upper()),
         pdf_name('/TransformParams'): generic.DictionaryObject({
             pdf_name('/Type'): pdf_name('/TransformParams'),
             pdf_name('/V'): pdf_name('/1.2'),
@@ -45,7 +45,6 @@ def docmdp_reference_dictionary(md_algorithm, permission_level: MDPPerm):
 
 
 def fieldmdp_reference_dictionary(field_mdp_spec: FieldMDPSpec,
-                                  md_algorithm: str,
                                   data_ref: generic.Reference):
     data_ref = generic.IndirectObject(
         data_ref.idnum, data_ref.generation, data_ref.pdf
@@ -55,7 +54,6 @@ def fieldmdp_reference_dictionary(field_mdp_spec: FieldMDPSpec,
         pdf_name('/Type'): pdf_name('/SigRef'),
         pdf_name('/TransformMethod'): pdf_name('/FieldMDP'),
         pdf_name('/Data'): data_ref,
-        pdf_name('/DigestMethod'): pdf_name('/' + md_algorithm.upper()),
         pdf_name('/TransformParams'): field_mdp_spec.as_transform_params()
     })
 
@@ -102,7 +100,8 @@ def _get_or_create_sigfield(field_name, pdf_out: BasePdfFileWriter,
                 'include_on_page': pdf_out.find_page_for_modification(
                     new_field_spec.on_page
                 )[0],
-                'combine_annotation': new_field_spec.combine_annotation
+                'combine_annotation': new_field_spec.combine_annotation,
+                'invis_settings': new_field_spec.invis_sig_settings,
             }
         else:
             sig_field_kwargs = {}
@@ -112,6 +111,11 @@ def _get_or_create_sigfield(field_name, pdf_out: BasePdfFileWriter,
             existing_fields_only=existing_fields_only,
             **sig_field_kwargs
         )
+        if field_created and new_field_spec is not None:
+            apply_sig_field_spec_properties(
+                pdf_out, sig_field=sig_field_ref.get_object(),
+                sig_field_spec=new_field_spec
+            )
 
     ensure_sig_flags(writer=pdf_out, lock_sig_flags=True)
 
@@ -162,7 +166,6 @@ class SigMDPSetup:
         docmdp_perms = self.docmdp_perms
 
         lock = self.field_lock
-        md_algorithm = self.md_algorithm
 
         reference_array = generic.ArrayObject()
 
@@ -178,12 +181,12 @@ class SigMDPSetup:
             perms[pdf_name('/DocMDP')] = sig_obj_ref
             writer.update_container(perms)
             reference_array.append(
-                docmdp_reference_dictionary(md_algorithm, docmdp_perms)
+                docmdp_reference_dictionary(docmdp_perms)
             )
 
         if lock is not None:
             fieldmdp_ref = fieldmdp_reference_dictionary(
-                lock, md_algorithm, data_ref=writer.root_ref
+                lock, data_ref=writer.root_ref
             )
             reference_array.append(fieldmdp_ref)
 
@@ -216,12 +219,12 @@ class SigAppearanceSetup:
     Timestamp to show in the signature appearance.
     """
 
-    name: str
+    name: Optional[str]
     """
     Signer name to show in the signature appearance.
     """
 
-    text_params: dict = None
+    text_params: Optional[dict] = None
     """
     Additional text interpolation parameters to pass to the underlying
     stamp style.
@@ -259,11 +262,11 @@ class SigAppearanceSetup:
 
         name = self.name
         timestamp = self.timestamp
-        extra_text_params = self.text_params or {}
-        text_params = {
-            'signer': name,
-            **extra_text_params
-        }
+        text_params = {}
+        if name is not None:
+            text_params['signer'] = name
+        text_params.update(self.text_params or {})
+
         if isinstance(style, TextStampStyle):
             text_params['ts'] = timestamp.strftime(style.timestamp_format)
 
