@@ -7,7 +7,7 @@ import logging
 import warnings
 from dataclasses import dataclass
 from datetime import datetime
-from typing import IO, Iterable, List, Optional, Union
+from typing import IO, Callable, Iterable, List, Optional, Union
 
 import tzlocal
 from asn1crypto import algos, cms, core, keys
@@ -28,7 +28,11 @@ from pyhanko_certvalidator._asyncio_compat import to_thread
 from pyhanko.pdf_utils import misc
 from pyhanko.sign import attributes
 from pyhanko.sign.ades.api import CAdESSignedAttrSpec
-from pyhanko.sign.attributes import CMSAttributeProvider
+from pyhanko.sign.attributes import (
+    CMSAttributeProvider,
+    SignedAttributeProviderSpec,
+    UnsignedAttributeProviderSpec,
+)
 from pyhanko.sign.general import (
     CertificateStore,
     SigningError,
@@ -57,12 +61,12 @@ logger = logging.getLogger(__name__)
 
 
 @dataclass(frozen=True)
-class PdfCMSSignedAttributes:
+class CMSSignedAttributes:
     """
-    .. versionadded:: 0.7.0
+    .. versionadded:: 0.14.0
 
     Serialisable container class describing input for various signed attributes
-    in a CMS object for a PDF signature.
+    in a signature CMS object.
     """
 
     signing_time: Optional[datetime] = None
@@ -71,17 +75,31 @@ class PdfCMSSignedAttributes:
     context.
     """
 
-    adobe_revinfo_attr: Optional[asn1_pdf.RevocationInfoArchival] = None
-    """
-    Adobe-style signed revocation info attribute.
-    """
-
     cades_signed_attrs: Optional[CAdESSignedAttrSpec] = None
     """
     Optional settings for CAdES-style signed attributes.
     """
 
 
+@dataclass(frozen=True)
+class PdfCMSSignedAttributes(CMSSignedAttributes):
+    """
+    .. versionadded:: 0.7.0
+
+    .. versionchanged:: 0.14.0
+        Split off some fields into :class:`.CMSSignedAttributes`.
+
+
+    Serialisable container class describing input for various signed attributes
+    in a CMS object for a PDF signature.
+    """
+
+    adobe_revinfo_attr: Optional[asn1_pdf.RevocationInfoArchival] = None
+    """
+    Adobe-style signed revocation info attribute.
+    """
+
+
 def _ensure_content_type(encap_content_info):
 
     encap_content_info = encap_content_info or {'content_type': 'data'}
@@ -252,9 +270,25 @@ class Signer:
             for most applications.
     """
 
-    signing_cert: x509.Certificate
+    signing_cert: Optional[x509.Certificate] = None
     """
+    .. versionchanged:: 0.14.0
+        Made optional (see note)
+
     The certificate that will be used to create the signature.
+
+    .. note::
+        This is an optional field only to a limited extent. Subclasses may
+        require it to be present, and not setting it at the beginning of
+        the signing process implies that certain high-level convenience
+        features will no longer work or be limited in function (e.g.,
+        automatic hash selection, appearance generation, revocation information
+        collection, ...).
+
+        However, making :attr:`signing_cert` optional enables certain signing
+        workflows where the certificate of the signer is not known until the
+        signature has actually been produced. This is most relevant in certain
+        types of remote signing scenarios.
     """
 
     cert_registry: CertificateStore
@@ -280,6 +314,9 @@ class Signer:
     def __init__(self, prefer_pss=False, embed_roots=True):
         self.prefer_pss = prefer_pss
         self.embed_roots = embed_roots
+        self.signed_attr_prov_spec: Optional[SignedAttributeProviderSpec] = None
+        self.unsigned_attr_prov_spec: Optional[UnsignedAttributeProviderSpec] \
+            = None
 
     def get_signature_mechanism(self, digest_algorithm):
         """
@@ -301,7 +338,7 @@ class Signer:
         if self.signing_cert is None:
             raise SigningError(
                 "Could not set up a default signature mechanism."
-            )  # pragma: nocover
+            )
         # Grab the certificate's algorithm (but forget about the digest)
         #  and use that to set up the default.
         # We'll specify the digest somewhere else.
@@ -349,14 +386,21 @@ class Signer:
         return SignedDigestAlgorithm(sda_kwargs)
 
     @property
-    def subject_name(self):
+    def subject_name(self) -> Optional[str]:
         """
         :return:
             The subject's common name as a string, extracted from
-            :attr:`signing_cert`.
+            :attr:`signing_cert`, or ``None`` if no signer's certificate is
+            available
         """
+        if self.signing_cert is None:
+            return None
+
         name: x509.Name = self.signing_cert.subject
-        result = name.native['common_name']
+        try:
+            result = name.native['common_name']
+        except KeyError:
+            result = name.native['organization_name']
         try:
             email = name.native['email_address']
             result = '%s <%s>' % (result, email)
@@ -386,6 +430,38 @@ class Signer:
         if revinfo_dict:
             return asn1_pdf.RevocationInfoArchival(revinfo_dict)
 
+    def _signed_attr_provider_spec(self,
+                                   attr_settings: PdfCMSSignedAttributes,
+                                   timestamper=None, use_cades=False,
+                                   is_pdf_sig=True):
+        """
+        Internal method to select a default attribute provider spec if none
+        is available already.
+        """
+
+        if self.signed_attr_prov_spec is not None:
+            return self.signed_attr_prov_spec
+        elif use_cades:
+            return CAdESSignedAttributeProviderSpec(
+                attr_settings=attr_settings,
+                signing_cert=self.signing_cert, is_pades=is_pdf_sig,
+                timestamper=timestamper
+            )
+        elif is_pdf_sig:
+            return GenericPdfSignedAttributeProviderSpec(
+                attr_settings=attr_settings,
+                signing_cert=self.signing_cert,
+                signature_mechanism=self.get_signature_mechanism,
+                timestamper=timestamper
+            )
+        else:
+            return GenericCMSSignedAttributeProviderSpec(
+                attr_settings=attr_settings,
+                signing_cert=self.signing_cert,
+                signature_mechanism=self.get_signature_mechanism,
+                timestamper=timestamper
+            )
+
     def _signed_attr_providers(self, data_digest: bytes, digest_algorithm: str,
                                attr_settings: PdfCMSSignedAttributes,
                                timestamper=None, use_cades=False,
@@ -393,51 +469,36 @@ class Signer:
         """
         Prepare "standard" signed attribute providers. Internal API.
         """
-        yield attributes.SigningCertificateV2Provider(
-            signing_cert=self.signing_cert
+
+        spec = self._signed_attr_provider_spec(
+            attr_settings=attr_settings,
+            timestamper=timestamper,
+            use_cades=use_cades,
+            is_pdf_sig=is_pdf_sig
         )
-        if not is_pdf_sig or not use_cades:
-            # NOTE: PAdES actually forbids this, but CAdES requires it!
-            signing_time = attr_settings.signing_time
-            if signing_time is None and not is_pdf_sig and use_cades:
-                # Ensure CAdES mandate is followed
-                signing_time = datetime.now(tz=tzlocal.get_localzone())
-            if signing_time is not None:
-                yield attributes.SigningTimeProvider(timestamp=signing_time)
-        if not use_cades:
-            if attr_settings.adobe_revinfo_attr is not None:
-                yield attributes.AdobeRevinfoProvider(
-                    value=attr_settings.adobe_revinfo_attr
-                )
 
-            # TODO not sure if PAdES/CAdES allow this, need to check.
-            #  It *should*, but perhaps the version of CMS it is based on is too
-            #  old, or it might not allow undefined signed attributes.
-            # In the meantime, we only add this attribute to non-PAdES sigs
-            yield attributes.CMSAlgorithmProtectionProvider(
-                digest_algo=digest_algorithm,
-                signature_algo=self.get_signature_mechanism(digest_algorithm)
-            )
-        cades_meta = attr_settings.cades_signed_attrs
-        # apply CAdES-specific attributes regardless of use_pades
-        if cades_meta is not None:
-            yield from cades_meta.prepare_providers(
-                message_digest=data_digest,
-                md_algorithm=digest_algorithm,
-                timestamper=timestamper
-            )
+        return spec.signed_attr_providers(
+            data_digest=data_digest, digest_algorithm=digest_algorithm,
+        )
 
-    def _unsigned_attr_providers(self, digest_algorithm, signature: bytes,
+    def _unsigned_attr_provider_spec(self,
+                                     timestamper: Optional[TimeStamper] = None):
+        if self.unsigned_attr_prov_spec is not None:
+            return self.unsigned_attr_prov_spec
+        else:
+            return DefaultUnsignedAttributes(timestamper=timestamper)
+
+    def _unsigned_attr_providers(self, digest_algorithm: str, signature: bytes,
+                                 signed_attrs: cms.CMSAttributes,
                                  timestamper: Optional[TimeStamper] = None):
         """
         Prepare "standard" unsigned attribute providers. Internal API.
         """
-        if timestamper is not None:
-            # the timestamp server needs to cross-sign our signature
-            yield attributes.TSTProvider(
-                digest_algorithm=digest_algorithm, data_to_ts=signature,
-                timestamper=timestamper
-            )
+        spec = self._unsigned_attr_provider_spec(timestamper)
+        return spec.unsigned_attr_providers(
+            digest_algorithm=digest_algorithm, signature=signature,
+            signed_attrs=signed_attrs
+        )
 
     def signer_info(self, digest_algorithm: str, signed_attrs, signature):
         """
@@ -457,6 +518,11 @@ class Signer:
         )
 
         signing_cert = self.signing_cert
+        if signing_cert is None:
+            raise SigningError(
+                "The signer\'s certificate must be available for "
+                "SignerInfo assembly to proceed."
+            )
         # build the signer info object that goes into the PKCS7 signature
         # (see RFC 2315 § 9.2)
         sig_info = cms.SignerInfo({
@@ -490,11 +556,12 @@ class Signer:
         if unsigned_attrs is not None:
             sig_info['unsigned_attrs'] = unsigned_attrs
 
+        cert_registry = self.cert_registry or ()
         # Note: we do not add the TS certs at this point
         if self.embed_roots:
             certs = {
                 cms.CertificateChoices(name='certificate', value=cert)
-                for cert in self.cert_registry
+                for cert in cert_registry
             }
         else:
             # asn1crypto's heuristic is good enough for now, we won't check the
@@ -504,7 +571,7 @@ class Signer:
             # asn1crypto either way.
             certs = {
                 cms.CertificateChoices(name='certificate', value=cert)
-                for cert in self.cert_registry if cert.self_signed == 'no'
+                for cert in cert_registry if cert.self_signed == 'no'
             }
         certs.add(
             cms.CertificateChoices(name='certificate', value=self.signing_cert)
@@ -567,13 +634,18 @@ class Signer:
         """
         raise NotImplementedError
 
-    async def unsigned_attrs(self, digest_algorithm, signature: bytes,
+    async def unsigned_attrs(self, digest_algorithm: str,
+                             signature: bytes,
+                             signed_attrs: cms.CMSAttributes,
                              timestamper=None, dry_run=False) \
             -> Optional[cms.CMSAttributes]:
         """
         .. versionchanged:: 0.9.0
             Made asynchronous _(breaking change)_
 
+        .. versionchanged:: 0.14.0
+            Added ``signed_attrs`` parameter _(breaking change)_
+
         Compute the unsigned attributes to embed into the CMS object.
         This function is called after signing the hash of the signed attributes
         (see :meth:`signed_attrs`).
@@ -586,6 +658,8 @@ class Signer:
 
         :param digest_algorithm:
             Digest algorithm used to hash the signed attributes.
+        :param signed_attrs:
+            Signed attributes of the signature.
         :param signature:
             Signature of the signed attribute hash.
         :param timestamper:
@@ -598,7 +672,9 @@ class Signer:
             The unsigned attributes to add, or ``None``.
         """
         provs = self._unsigned_attr_providers(
-            digest_algorithm=digest_algorithm, signature=signature,
+            signature=signature,
+            signed_attrs=signed_attrs,
+            digest_algorithm=digest_algorithm,
             timestamper=timestamper
         )
         attrs = await format_attributes(list(provs), dry_run=dry_run)
@@ -793,7 +869,9 @@ class Signer:
             signed_attrs.dump(), digest_algorithm.lower(), dry_run
         )
         unsigned_attrs = await self.unsigned_attrs(
-            digest_algorithm, signature, timestamper=timestamper,
+            digest_algorithm, signature,
+            signed_attrs=signed_attrs,
+            timestamper=timestamper,
             dry_run=dry_run
         )
         return self._package_signature(
@@ -1335,17 +1413,36 @@ class ExternalSigner(Signer):
     It embeds a fixed signature value into the CMS, set at initialisation.
 
     Intended for use with :ref:`interrupted-signing`.
+
+    :param signing_cert:
+        The signer's certificate.
+    :param cert_registry:
+        The certificate registry to use in CMS generation.
+    :param signature_value:
+        The value of the signature as a byte string, a placeholder length,
+        or ``None``.
+    :param signature_mechanism:
+        The signature mechanism used by the external signing service.
+    :param prefer_pss:
+        Switch to prefer PSS when producing RSA signatures, as opposed to
+        RSA with PKCS#1 v1.5 padding.
+    :param embed_roots:
+        Whether to embed relevant root certificates into the CMS payload.
     """
 
-    def __init__(self, signing_cert: x509.Certificate,
-                 cert_registry: CertificateStore,
-                 signature_value: bytes,
-                 signature_mechanism: SignedDigestAlgorithm = None,
-                 prefer_pss=False, embed_roots=True):
+    def __init__(self, signing_cert: Optional[x509.Certificate],
+                 cert_registry: Optional[CertificateStore],
+                 signature_value: Union[bytes, int, None] = None,
+                 signature_mechanism: Optional[SignedDigestAlgorithm] = None,
+                 prefer_pss: bool = False, embed_roots: bool = True):
         self.signing_cert = signing_cert
-        self.cert_registry = cert_registry
+        self.cert_registry = cert_registry or SimpleCertificateStore()
         self.signature_mechanism = signature_mechanism
-        self._signature_value = signature_value
+
+        if isinstance(signature_value, bytes):
+            self._signature_value = signature_value
+        else:
+            self._signature_value = bytes(signature_value or 256)
         super().__init__(prefer_pss=prefer_pss, embed_roots=embed_roots)
 
     async def async_sign_raw(self,
@@ -1357,34 +1454,142 @@ class ExternalSigner(Signer):
         return self._signature_value
 
 
-# TODO consider deprecating the current signed_attrs kwargs in favour of this
-#  dataclass
-
-@dataclass(frozen=True)
-class PdfCMSSignedAttributes:
+class GenericCMSSignedAttributeProviderSpec(SignedAttributeProviderSpec):
     """
-    .. versionadded:: 0.7.0
-
-    Serialisable container class describing input for various signed attributes
-    in a CMS object for a PDF signature.
+    Signed attribute provider spec for generic CMS signatures.
     """
 
-    signing_time: Optional[datetime] = None
+    def __init__(self,
+                 attr_settings: CMSSignedAttributes,
+                 signing_cert: Optional[x509.Certificate],
+                 signature_mechanism: Union[
+                     Callable[[str], algos.SignedDigestAlgorithm],
+                     None
+                 ],
+                 timestamper: Optional[TimeStamper]):
+        self.signing_cert = signing_cert
+        self.attr_settings = attr_settings
+        self.signature_mechanism = signature_mechanism
+        self.timestamper = timestamper
+
+    def signed_attr_providers(self, data_digest: bytes, digest_algorithm: str) \
+            -> Iterable[CMSAttributeProvider]:
+        attr_settings = self.attr_settings
+        if self.signing_cert is not None:
+            yield attributes.SigningCertificateV2Provider(
+                signing_cert=self.signing_cert
+            )
+        signing_time = attr_settings.signing_time
+        if signing_time is not None:
+            yield attributes.SigningTimeProvider(timestamp=signing_time)
+        if attr_settings.cades_signed_attrs is not None:
+            yield from attr_settings.cades_signed_attrs.prepare_providers(
+                message_digest=data_digest,
+                md_algorithm=digest_algorithm,
+                timestamper=self.timestamper
+            )
+
+        if self.signature_mechanism is not None:
+            mech = self.signature_mechanism(digest_algorithm)
+            # TODO not sure if PAdES/CAdES allow this, need to check.
+            #  It *should*, but perhaps the version of CMS it is based on is too
+            #  old, or it might not allow undefined signed attributes.
+            # In the meantime, we only add this attribute to non-PAdES sigs
+            yield attributes.CMSAlgorithmProtectionProvider(
+                digest_algo=digest_algorithm,
+                signature_algo=mech
+            )
+
+
+class GenericPdfSignedAttributeProviderSpec(GenericCMSSignedAttributeProviderSpec):
     """
-    Timestamp for the ``signingTime`` attribute. Will be ignored in a PAdES
-    context.
+    Signed attribute provider spec for generic PDF signatures.
     """
 
-    adobe_revinfo_attr: Optional[cms.CMSAttribute] = None
+    def __init__(self,
+                 attr_settings: PdfCMSSignedAttributes,
+                 signing_cert: Optional[x509.Certificate],
+                 signature_mechanism: Union[
+                     Callable[[str], algos.SignedDigestAlgorithm],
+                     None
+                 ],
+                 timestamper: Optional[TimeStamper]):
+        super().__init__(
+            attr_settings=attr_settings,
+            signing_cert=signing_cert,
+            signature_mechanism=signature_mechanism,
+            timestamper=timestamper
+        )
+
+    def signed_attr_providers(self, data_digest: bytes, digest_algorithm: str) \
+            -> Iterable[CMSAttributeProvider]:
+        yield from super().signed_attr_providers(
+            data_digest=data_digest, digest_algorithm=digest_algorithm
+
+        )
+        attr_settings = self.attr_settings
+        assert isinstance(attr_settings, PdfCMSSignedAttributes)
+        if attr_settings.adobe_revinfo_attr is not None:
+            yield attributes.AdobeRevinfoProvider(
+                value=attr_settings.adobe_revinfo_attr)
+
+
+class CAdESSignedAttributeProviderSpec(SignedAttributeProviderSpec):
     """
-    Adobe-style signed revocation info attribute.
+    Signed attribute provider spec for CAdES and PAdES signatures.
     """
 
-    cades_signed_attrs: Optional[CAdESSignedAttrSpec] = None
+    def __init__(self, attr_settings: CMSSignedAttributes,
+                 signing_cert: x509.Certificate, is_pades: bool,
+                 timestamper: Optional[TimeStamper]):
+        self.signing_cert = signing_cert
+        self.attr_settings = attr_settings
+        self.is_pades = is_pades
+        self.timestamper = timestamper
+
+    def signed_attr_providers(self, data_digest: bytes, digest_algorithm: str) \
+            -> Iterable[CMSAttributeProvider]:
+
+        yield attributes.SigningCertificateV2Provider(
+            signing_cert=self.signing_cert
+        )
+        attr_settings = self.attr_settings
+        if not self.is_pades:
+            # NOTE: PAdES actually forbids this, but CAdES requires it!
+            signing_time = attr_settings.signing_time
+            if signing_time is None:
+                # Ensure CAdES mandate is followed
+                signing_time = datetime.now(tz=tzlocal.get_localzone())
+            if signing_time is not None:
+                yield attributes.SigningTimeProvider(timestamp=signing_time)
+        cades_meta = attr_settings.cades_signed_attrs
+        if cades_meta is not None:
+            yield from cades_meta.prepare_providers(
+                message_digest=data_digest,
+                md_algorithm=digest_algorithm,
+                timestamper=self.timestamper
+            )
+
+
+class DefaultUnsignedAttributes(UnsignedAttributeProviderSpec):
     """
-    Optional settings for CAdES-style signed attributes.
+    Default unsigned attribute provider spec.
     """
 
+    def __init__(self, timestamper: Optional[TimeStamper]):
+        self.timestamper = timestamper
+
+    def unsigned_attr_providers(
+            self, signature: bytes, signed_attrs: cms.CMSAttributes,
+            digest_algorithm: str) -> Iterable[CMSAttributeProvider]:
+        timestamper = self.timestamper
+        if timestamper is not None:
+            # the timestamp server needs to cross-sign our signature
+            yield attributes.TSTProvider(
+                digest_algorithm=digest_algorithm, data_to_ts=signature,
+                timestamper=timestamper
+            )
+
 
 RSA_THRESHOLDS = [(2048, 'sha256'), (3072, 'sha384')]
 ECC_THRESHOLDS = [(256, 'sha256'), (384, 'sha384')]
