@@ -27,6 +27,7 @@ from pyhanko.pdf_utils.writer import BasePdfFileWriter
 from pyhanko.sign.ades.api import CAdESSignedAttrSpec
 from pyhanko.sign.fields import (
     FieldMDPSpec,
+    InvisSigSettings,
     MDPPerm,
     SeedLockDocument,
     SigFieldSpec,
@@ -63,6 +64,7 @@ __all__ = [
     'PreSignValidationStatus', 'PostSignInstructions'
 ]
 
+from ...pdf_utils.crypt import SerialisedCredential
 
 logger = logging.getLogger(__name__)
 
@@ -437,9 +439,13 @@ class PdfTimeStamper:
     """
 
     def __init__(self, timestamper: TimeStamper,
-                 field_name: Optional[str] = None):
+                 field_name: Optional[str] = None,
+                 invis_settings: InvisSigSettings = InvisSigSettings(),
+                 readable_field_name: str = "Timestamp"):
         self.default_timestamper = timestamper
         self._field_name = field_name
+        self._readable_field_name = readable_field_name
+        self._invis_settings = invis_settings
 
     @property
     def field_name(self) -> str:
@@ -635,7 +641,12 @@ class PdfTimeStamper:
 
         timestamp_obj = DocumentTimestamp(bytes_reserved=bytes_reserved)
 
-        cms_writer = PdfCMSEmbedder().write_cms(
+        field_spec = SigFieldSpec(
+            sig_field_name=field_name,
+            invis_sig_settings=self._invis_settings,
+            readable_field_name=self._readable_field_name,
+        )
+        cms_writer = PdfCMSEmbedder(new_field_spec=field_spec).write_cms(
             field_name=field_name, writer=pdf_out,
             # for LTA, requiring existing_fields_only doesn't make sense
             # since we should in principle be able to add document timestamps
@@ -665,7 +676,14 @@ class PdfTimeStamper:
                 output_stream=res_output, sig_contents=sig_contents,
                 paths=validation_paths, validation_context=validation_context,
                 force_write=not dss_settings.skip_if_unneeded,
-                embed_roots=embed_roots
+                embed_roots=embed_roots,
+                # FIXME in this case, the ser/deser step is unnecessary
+                #  and inefficient; should probably rewrite
+                #  using supply_dss_in_writer
+                file_credential=(
+                    pdf_out.security_handler.extract_credential().serialise()
+                    if pdf_out.security_handler else None
+                )
             )
 
         return misc.finalise_output(output, res_output)
@@ -938,10 +956,17 @@ class PdfSigner:
             md_algorithm = self.default_md_for_signer
         elif sv_md_algorithm is not None:
             md_algorithm = sv_md_algorithm
-        else:
+        elif self.signer.signing_cert is not None:
             md_algorithm = select_suitable_signing_md(
                 self.signer.signing_cert.public_key
             )
+        else:
+            raise SigningError(
+                "Could not select a default digest algorithm. Please supply "
+                "a value in the signature settings, or configure the signer "
+                "with an explicit signature mechanism that includes a digest "
+                "algorithm specification."
+            )
 
         # TODO fall back to more useful default for weak_hash_algos
         weak_hash_algos = (
@@ -980,6 +1005,15 @@ class PdfSigner:
             in its post-setup stage.
         """
 
+        if isinstance(pdf_out, IncrementalPdfFileWriter):
+            # ensure we're not signing a hybrid reference doc
+            prev = pdf_out.prev
+            if prev.strict and prev.xrefs.hybrid_xrefs_present:
+                raise SigningError(
+                    "Attempting to sign document with hybrid cross-reference "
+                    "sections while hybrid xrefs are disabled"
+                )
+
         timestamper = self.default_timestamper
 
         signature_meta: PdfSignatureMetadata = self.signature_meta
@@ -1061,7 +1095,7 @@ class PdfSigner:
         :param bytes_reserved:
             Bytes to reserve for the CMS object in the PDF file.
             If not specified, make an estimate based on a dummy signature.
-            
+
             .. warning::
                 Since the CMS object is written to the output file as a
                 hexadecimal string, you should request **twice** the (estimated)
@@ -1408,11 +1442,17 @@ class PdfSigningSession:
         validation_context = signature_meta.validation_context
 
         if signature_meta.embed_validation_info:
-            if validation_context is None:
+            if self.pdf_signer.signer.signing_cert is None:
                 raise SigningError(
-                    'A validation context must be provided if '
-                    'validation/revocation info is to be embedded into the '
-                    'signature.'
+                    "A signer's certificate must be provided if "
+                    "validation/revocation info is to be embedded into the "
+                    "signature."
+                )
+            elif validation_context is None:
+                raise SigningError(
+                    "A validation context must be provided if "
+                    "validation/revocation info is to be embedded into the "
+                    "signature."
                 )
             elif not validation_context.fetching_allowed:
                 logger.warning(
@@ -1656,6 +1696,11 @@ class PdfSigningSession:
         flags: SigSeedValFlags = sv_spec.flags
 
         if sv_spec.cert is not None:
+            if pdf_signer.signer.signing_cert is None:
+                raise SigningError(
+                    "Cannot verify seed value constraints on the signer's "
+                    "certificate since it is not available"
+                )
             sv_spec.cert.satisfied_by(
                 pdf_signer.signer.signing_cert, validation_path
             )
@@ -1769,6 +1814,13 @@ class PdfSigningSession:
         md_algorithm = self.md_algorithm
         signature_meta = self.pdf_signer.signature_meta
         signer = self.pdf_signer.signer
+
+        if signer.signing_cert is None:
+            raise SigningError(
+                "Automatic signature size estimation is not available without "
+                "a signer's certificate. Space must be allocated manually "
+                "using bytes_reserved=..."
+            )
         # estimate bytes_reserved by creating a fake CMS object
         md_spec = get_pyca_cryptography_hash(md_algorithm)
         test_md = hashes.Hash(md_spec).finalize()
@@ -1885,6 +1937,12 @@ class PdfSigningSession:
         if self.use_pades and signature_meta.embed_validation_info:
             if signature_meta.use_pades_lta:
                 doc_timestamper = self.timestamper
+            # if necessary/supported, extract a file access credential
+            # to perform post-signing operations later
+            if self.pdf_out.security_handler is not None:
+                credential = self.pdf_out.security_handler.extract_credential()
+            else:
+                credential = None
             post_signing_instr = PostSignInstructions(
                 validation_info=validation_info,
                 # use the same algorithm
@@ -1897,7 +1955,8 @@ class PdfSigningSession:
                 timestamp_field_name=signature_meta.timestamp_field_name,
                 dss_settings=signature_meta.dss_settings,
                 tight_size_estimates=signature_meta.tight_size_estimates,
-                embed_roots=embed_roots
+                embed_roots=embed_roots,
+                file_credential=credential
             )
         return PdfTBSDocument(
             cms_writer=self.cms_writer, signer=pdf_signer.signer,
@@ -1980,6 +2039,13 @@ class PostSignInstructions:
         :class:`.Signer` involved, not from the initial configuration.
     """
 
+    file_credential: Optional[SerialisedCredential] = None
+    """
+    .. versionadded:: 0.13.0
+
+    Serialised file credential, to update encrypted files.
+    """
+
 
 class PdfTBSDocument:
     """
@@ -2274,11 +2340,18 @@ class PdfPostSignatureDocument:
             if not dss_settings.skip_if_unneeded:
                 dss_op_kwargs['force_write'] = True
             validation.DocumentSecurityStore.add_dss(
-                output_stream=output, **dss_op_kwargs
+                output_stream=output, **dss_op_kwargs,
+                file_credential=instr.file_credential
             )
         if timestamper is not None:
             # append a document timestamp after the DSS update
             w = IncrementalPdfFileWriter(output)
+            if w.security_handler is not None \
+                    and self.post_sign_instructions.file_credential is not None:
+                w.security_handler.authenticate(
+                    self.post_sign_instructions.file_credential
+                )
+                # we let the SH throw errors on access as necessary
             pdf_timestamper = PdfTimeStamper(
                 timestamper, field_name=instr.timestamp_field_name
             )
